import psutil
import joblib
import itertools
import io
import time
import tensorflow as tf
from tensorboard.plugins.hparams import api as hp
import os
import numpy as np
import gc
import time
import matplotlib.pyplot as plt
import json
import logging
from pathlib import Path

from sklearn.metrics import confusion_matrix
import cv2
from ml_tools import tools
from ml_tools.datasetstructures import SegmentType

from ml_tools.preprocess import (
    preprocess_movement,
    preprocess_frame,
)
from ml_tools.interpreter import Interpreter
from classify.trackprediction import TrackPrediction
from ml_tools.hyperparams import HyperParams
from ml_tools import thermaldataset
from ml_tools.resnet.wr_resnet import WRResNet

from ml_tools import irdataset
from ml_tools.tfdataset import get_weighting, get_distribution, get_dataset as get_tf
from ml_tools import forestmodel
from ml_tools.preprocess import FrameTypes

classify_i = 0


class KerasModel(Interpreter):
    """Defines a deep learning model"""

    VERSION = 1
    TYPE = "Keras"

    def __init__(self, train_config=None, labels=None, data_dir=None):
        self.model = None
        self.datasets = None
        self.remapped = None
        # dictionary containing current hyper parameters
        self.params = HyperParams()
        self.data_type = None
        self.data_dir = data_dir
        if train_config:
            self.log_base = train_config.train_dir / "logs"
            self.log_dir = self.log_base
            self.checkpoint_folder = train_config.train_dir / "checkpoints"
            self.params.update(train_config.hyper_params)
            self.data_type = train_config.type
        self.labels = labels
        self.preprocess_fn = None
        self.validate = None
        self.train = None
        self.test = None
        self.mapped_labels = None
        self.label_probabilities = None
        self.class_weights = None
        self.ds_by_label = True
        self.excluded_labels = None
        self.remapped_labels = None
        self.orig_labels = None

    def load_training_meta(self, base_dir):
        file = f"{base_dir}/training-meta.json"
        logging.info("loading meta %s", file)
        with open(file, "r") as f:
            meta = json.load(f)
        self.labels = meta.get("labels", [])
        self.data_type = meta.get("type", "thermal")
        self.dataset_counts = meta.get("counts")
        self.ds_by_label = meta.get("by_label", True)
        self.excluded_labels = meta.get("excluded_labels")
        self.remapped_labels = meta.get("remapped_labels")
        self.params.set_use_segments(
            meta.get("config").get("build", {}).get("use_segments", True)
        )

    def shape(self):
        if self.model is None:
            return None
        inputs = self.model.inputs
        shape = []
        for input in inputs:
            shape.append(tuple(input.shape.as_list()))
        return len(shape), shape

    def get_base_model(self, input, weights="imagenet"):
        pretrained_model = self.params.model_name
        if pretrained_model == "resnet":
            return (
                tf.keras.applications.ResNet50(
                    weights=weights,
                    include_top=False,
                    input_tensor=input,
                ),
                tf.keras.applications.resnet.preprocess_input,
            )
        elif pretrained_model == "resnetv2":
            return (
                tf.keras.applications.ResNet50V2(
                    weights=weights, include_top=False, input_tensor=input
                ),
                tf.keras.applications.resnet_v2.preprocess_input,
            )
        elif pretrained_model == "resnet152":
            return (
                tf.keras.applications.ResNet152(
                    weights=weights, include_top=False, input_tensor=input
                ),
                tf.keras.applications.resnet.preprocess_input,
            )
        elif pretrained_model == "vgg16":
            return (
                tf.keras.applications.VGG16(
                    weights=weights,
                    include_top=False,
                    input_tensor=input,
                ),
                tf.keras.applications.vgg16.preprocess_input,
            )
        elif pretrained_model == "vgg19":
            return (
                tf.keras.applications.VGG19(
                    weights=weights,
                    include_top=False,
                    input_tensor=input,
                ),
                tf.keras.applications.vgg19.preprocess_input,
            )
        elif pretrained_model == "mobilenet":
            return (
                tf.keras.applications.MobileNetV2(
                    weights=weights,
                    include_top=False,
                    input_tensor=input,
                ),
                tf.keras.applications.mobilenet_v2.preprocess_input,
            )
        elif pretrained_model == "densenet121":
            return (
                tf.keras.applications.DenseNet121(
                    weights=weights,
                    include_top=False,
                    input_tensor=input,
                ),
                tf.keras.applications.densenet.preprocess_input,
            )
        elif pretrained_model == "inceptionresnetv2":
            return (
                tf.keras.applications.InceptionResNetV2(
                    weights=weights,
                    include_top=False,
                    input_tensor=input,
                ),
                tf.keras.applications.inception_resnet_v2.preprocess_input,
            )
        elif pretrained_model == "inceptionv3":
            print("Input", input)
            return (
                tf.keras.applications.InceptionV3(
                    weights=weights,
                    include_top=False,
                    input_tensor=input,
                ),
                tf.keras.applications.inception_v3.preprocess_input,
            )
        elif pretrained_model == "efficientnetb5":
            return (
                tf.keras.applications.EfficientNetB5(
                    weights=weights,
                    include_top=False,
                    input_tensor=input,
                ),
                None,
            )
        elif pretrained_model == "efficientnetb0":
            return (
                tf.keras.applications.EfficientNetB0(
                    weights=weights,
                    include_top=False,
                    input_tensor=input,
                ),
                None,
            )
        elif pretrained_model == "efficientnetb1":
            return (
                tf.keras.applications.EfficientNetB1(
                    weights=weights,
                    include_top=False,
                    input_tensor=input,
                ),
                None,
            )
        elif pretrained_model == "nasnet":
            return (
                tf.keras.applications.nasnet.NASNetLarge(
                    weights=weights, include_top=False, input_tensor=input
                ),
                tf.keras.applications.nasnet.preprocess_input,
            )
        elif pretrained_model == "wr-resnet":
            return (
                WRResNet(
                    input, k=self.params.get("k", 4), depth=self.params.get("depth", 22)
                ),
                None,
            )
        raise Exception("Could not find model " + pretrained_model)

    def get_forest_model(self, run_name):
        import tensorflow_decision_forests as tfdf

        train_files = self.data_dir / "train"
        train, remapped, _, _ = get_dataset(
            train_files,
            self.data_type,
            self.orig_labels,
            batch_size=self.params.batch_size,
            image_size=self.params.output_dim[:2],
            preprocess_fn=self.preprocess_fn,
            augment=False,
            resample=False,
            stop_on_empty_dataset=False,
            only_features=True,
            one_hot=False,
            excluded_labels=self.excluded_labels,
            remapped_labels=self.remapped_labels,
        )
        # have to run fit firest
        rf = tfdf.keras.RandomForestModel()
        rf.fit(train)
        rf.save(str(self.checkpoint_folder / run_name / "rf"))
        save_metadata(self)
        return rf

    def build_model(
        self, dense_sizes=None, retrain_from=None, dropout=None, run_name=None
    ):
        # width = self.params.frame_size
        width = self.params.output_dim[0]
        inputs = tf.keras.Input(
            shape=(width, width, len(self.params.channels)), name="input"
        )
        weights = None if self.params.base_training else "imagenet"
        base_model, preprocess = self.get_base_model(inputs, weights=weights)
        self.preprocess_fn = preprocess

        # inputs = base_model.input
        x = base_model.output
        # x = base_model(inputs, training=self.params.base_training)
        if self.params.get("model_merge"):
            logging.info(
                "Loading cnn rf model %s %s",
                self.params.get("model_cnn"),
                self.params.get("model_rf"),
            )
            cnn = tf.keras.models.load_model(self.params.get("model_cnn"))
            cnn.load_weights(
                Path(self.params.get("model_cnn")) / "val_acc"
            ).expect_partial()
            feature_input = tf.keras.Input(shape=(188), name="feature_input")
            model_rf = tf.keras.models.load_model(self.params.get("model_rf"))
            rf = model_rf(feature_input)
            inputs = [cnn.input, feature_input]
            cnn.summary()
            model_rf.summary()
            print("Outputs", cnn.outputs, rf)
            x = tf.keras.layers.Concatenate()([cnn.outputs[0], rf])
            activation = "softmax"
            if self.params.multi_label:
                activation = "sigmoid"
            logging.info("Using %s activation", activation)
            preds = tf.keras.layers.Dense(
                len(self.labels), activation=activation, name="merged-prediction"
            )(x)
            self.model = tf.keras.models.Model(inputs, outputs=preds)
        elif self.params.lstm:
            x = tf.keras.layers.GlobalAveragePooling2D()(x)
            for i in dense_sizes:
                x = tf.keras.layers.Dense(i, activation="relu")(x)
            # gp not sure how many should be pre lstm, and how many post
            cnn = tf.keras.models.Model(inputs, outputs=x)

            self.model = self.add_lstm(cnn)
        else:
            x = tf.keras.layers.GlobalAveragePooling2D()(x)
            if self.params.mvm:
                mvm_inputs = tf.keras.layers.Input((188))
                inputs = [inputs, mvm_inputs]
                # mvm_features = tf.keras.layers.Flatten()(mvm_inputs)
                #
                # if self.params["hq_mvm"]:
                # print("HQ")
                if self.params.mvm_forest:
                    rf = self.get_forest_model(run_name)

                    rf = rf(mvm_inputs)
                    x = tf.keras.layers.Concatenate()([x, rf])

                else:
                    mvm_features = tf.keras.layers.Dense(128, activation="relu")(
                        mvm_inputs
                    )
                    mvm_features = tf.keras.layers.Dense(128, activation="relu")(
                        mvm_features
                    )
                    mvm_features = tf.keras.layers.Dropout(0.1)(mvm_features)

                    # else:
                    #     mvm_features = tf.keras.layers.Dense(32, activation="relu")(
                    #         mvm_inputs
                    #     )
                    x = tf.keras.layers.Concatenate()([x, mvm_features])
                # x = tf.keras.layers.Dense(1028, activation="relu")(x)
            if dense_sizes is not None:
                for i in dense_sizes:
                    x = tf.keras.layers.Dense(i, activation="relu")(x)
            if dropout:
                x = tf.keras.layers.Dropout(dropout)(x)

            activation = "softmax"
            if self.params.multi_label:
                activation = "sigmoid"
            logging.info("Using %s activation", activation)
            preds = tf.keras.layers.Dense(
                len(self.labels), activation=activation, name="prediction"
            )(x)
            self.model = tf.keras.models.Model(inputs, outputs=preds)
        if retrain_from is None:
            retrain_from = self.params.retrain_layer
        if retrain_from:
            for i, layer in enumerate(base_model.layers):
                if isinstance(layer, tf.keras.layers.BatchNormalization):
                    # apparently this shouldn't matter as we set base_training = False
                    layer.trainable = False
                    logging.info("dont train %s %s", i, layer.name)
                else:
                    layer.trainable = i >= retrain_from
        else:
            base_model.trainable = self.params.base_training

        if self.params.multi_label:
            acc = tf.metrics.binary_accuracy
        else:
            acc = tf.metrics.categorical_accuracy

        self.model.compile(
            optimizer=optimizer(self.params),
            loss=loss(self.params),
            metrics=[
                acc,
                tf.keras.metrics.AUC(),
                tf.keras.metrics.Recall(),
                tf.keras.metrics.Precision(),
            ],
        )

    def adjust_final_layer(self):
        # Adjust final layer to a new set of labels, by removing it and re adding
        # new_model = tf.keras.models.Sequential(self.model.layers[:-3])
        self.model = tf.keras.Model(
            inputs=self.model.input, outputs=self.model.layers[-2].output
        )

        # model = tf.keras.Model(inputs=self.model.input, outputs=x)

        activation = "softmax"
        if self.params.multi_label:
            activation = "sigmoid"

        retrain_from = self.params.retrain_layer
        if retrain_from:
            for i, layer in enumerate(self.model.layers):
                if isinstance(layer, tf.keras.layers.BatchNormalization):
                    # apparently this shouldn't matter as we set base_training = False
                    layer.trainable = False
                    logging.info("dont train %s %s", i, layer.name)
                else:
                    layer.trainable = i >= retrain_from
        else:
            self.model.trainable = self.params.base_training

        # add final layer after as always want this trainable
        logging.info(
            "Adding new final layer with %s activation and %s labels ",
            activation,
            len(self.labels),
        )
        preds = tf.keras.layers.Dense(
            len(self.labels), activation=activation, name="prediction"
        )(self.model.output)

        self.model = tf.keras.models.Model(self.model.inputs, outputs=preds)
        if self.params.multi_label:
            acc = tf.metrics.binary_accuracy
        else:
            acc = tf.metrics.categorical_accuracy
        logging.info("Using acc %s", acc)
        self.model.summary()
        self.model.compile(
            optimizer=optimizer(self.params),
            loss=loss(self.params),
            metrics=[
                acc,
                tf.keras.metrics.AUC(),
                tf.keras.metrics.Recall(),
                tf.keras.metrics.Precision(),
            ],
        )

    def load_model(self, model_file, training=False, weights=None):
        model_file = Path(model_file)
        super().__init__(model_file)
        logging.info("Loading %s with model weights %s", model_file, weights)
        if model_file.suffix == ".pb":
            self.model = tf.keras.models.load_model(model_file.parent)
        else:
            self.model = tf.keras.models.load_model(model_file)

        self.model.trainable = training

        if weights is not None:
            self.model.load_weights(weights)
            logging.info("Loaded weight %s", weights)
        # print(self.model.summary())

    def save(self, run_name=None, history=None, test_results=None):
        # create a save point
        if run_name is None:
            run_name = self.params.model_name

        self.model.save(str(self.checkpoint_folder / run_name / f"{run_name}.keras"))
        self.save_metadata(run_name, history, test_results)

    def save_metadata(self, run_name=None, history=None, test_results=None):
        #  save metadata
        if run_name is None:
            run_name = self.params.model_name
        model_stats = {}
        model_stats["name"] = self.params.model_name
        model_stats["labels"] = self.labels
        model_stats["hyperparams"] = self.params
        model_stats["training_date"] = str(time.time())
        model_stats["version"] = self.VERSION
        model_stats["mapped_labels"] = self.mapped_labels
        model_stats["label_probabilities"] = self.label_probabilities
        model_stats["type"] = self.data_type
        model_stats["remapped_labels"] = self.remapped_labels
        model_stats["excluded_labels"] = self.excluded_labels
        if self.remapped is not None:
            model_stats["remapped"] = self.remapped
        if self.class_weights is not None:
            model_stats["class_weights"] = self.class_weights

        if history:
            json_history = {}
            for key, item in history.items():
                if isinstance(item, list) and isinstance(item[0], np.floating):
                    json_history[key] = [float(i) for i in item]
                else:
                    json_history[key] = item
            model_stats["history"] = json_history
        if test_results:
            model_stats["test_loss"] = test_results[0]
            model_stats["test_acc"] = test_results[1]
        run_dir = self.checkpoint_folder / run_name
        run_dir.mkdir(parents=True, exist_ok=True)
        if not run_dir.exists:
            run_dir.mkdir()

        json.dump(
            model_stats,
            (run_dir / f"{run_name}.json").open("w"),
            indent=4,
            cls=MetaJSONEncoder,
        )

    def close(self):
        # if self.test:
        #     self.test.stop_load()
        # if self.validate:
        #     self.validate.stop_load()
        # if self.train:
        #     self.train.stop_load()

        self.validate = None
        self.test = None
        self.train = None
        self.model = None
        tf.keras.backend.clear_session()
        gc.collect()
        del self.model
        del self.train
        del self.validate
        del self.test
        gc.collect()

    def train_model(
        self,
        epochs,
        run_name,
        weights=None,
        rebalance=False,
        resample=False,
        fine_tune=None,
    ):
        logging.info(
            "%s Training model for %s epochs with weights %s", run_name, epochs, weights
        )
        if self.params.excluded_labels is not None:
            self.excluded_labels = self.params.excluded_labels
        else:
            self.excluded_labels, self.remapped_labels = get_excluded(
                self.data_type, self.params.multi_label
            )
        if self.params.remapped_labels is not None:
            self.remapped_labels = self.params.remapped_labels
        else:
            self.remapped_labels, self.remapped_labels = get_excluded(
                self.data_type, self.params.multi_label
            )
        train_files = self.data_dir / "train"
        validate_files = self.data_dir / "validation"
        logging.info(
            "Excluding %s remapping %s", self.excluded_labels, self.remapped_labels
        )

        if self.params.multi_label and "land-bird" not in self.labels:
            self.labels.append("land-bird")
        self.orig_labels = self.labels.copy()

        self.preprocess_fn = self.get_preprocess_fn()
        self.train, remapped, new_labels, epoch_size = get_dataset(
            train_files,
            self.data_type,
            self.orig_labels,
            batch_size=self.params.batch_size,
            image_size=self.params.output_dim[:2],
            preprocess_fn=self.preprocess_fn,
            resample=resample,
            stop_on_empty_dataset=False,
            include_features=self.params.mvm,
            augment=True,
            excluded_labels=self.excluded_labels,
            remapped_labels=self.remapped_labels,
            # dist=self.dataset_counts["train"],
            multi_label=self.params.multi_label,
            num_frames=self.params.square_width**2,
            channels=self.params.channels,
        )
        self.labels = new_labels

        self.log_dir = self.log_base / run_name
        self.log_dir.mkdir(parents=True, exist_ok=True)
        if fine_tune is not None:
            self.load_model(fine_tune, weights=weights, training=True)
            # load model loads old labels
            self.labels = new_labels

            self.adjust_final_layer()
        else:

            if not self.model:
                self.build_model(
                    dense_sizes=self.params.dense_sizes,
                    retrain_from=self.params.retrain_layer,
                    dropout=self.params.dropout,
                    run_name=run_name,
                )

            if weights is not None:
                self.model.load_weights(weights)

        self.model.summary()

        self.remapped = remapped
        self.validate, remapped, _, _ = get_dataset(
            validate_files,
            self.data_type,
            self.orig_labels,
            batch_size=self.params.batch_size,
            image_size=self.params.output_dim[:2],
            preprocess_fn=self.preprocess_fn,
            resample=resample,
            stop_on_empty_dataset=False,
            include_features=self.params.mvm,
            excluded_labels=self.excluded_labels,
            remapped_labels=self.remapped_labels,
            multi_label=self.params.multi_label,
            num_frames=self.params.square_width**2,
            channels=self.params.channels,
        )
        if rebalance:
            self.class_weights = get_weighting(self.train, self.labels)
        logging.info(
            "Training on %s  with class weights %s",
            self.labels,
            self.class_weights,
        )

        self.save_metadata(run_name)
        self.save(run_name)
        checkpoints = self.checkpoints(run_name)

        history = self.model.fit(
            self.train,
            validation_data=self.validate,
            epochs=epochs,
            shuffle=False,
            class_weight=self.class_weights,
            callbacks=[
                tf.keras.callbacks.TensorBoard(
                    self.log_dir, write_graph=True, write_images=True
                ),
                *checkpoints,
            ],  # log metricslast_stats
        )
        history = history.history
        test_accuracy = None
        test_files = self.data_dir / "test"

        if len(list(test_files.glob("*.tfrecord"))) > 0:
            self.test, _, _, _ = get_dataset(
                test_files,
                self.data_type,
                self.orig_labels,
                batch_size=self.params.batch_size,
                image_size=self.params.output_dim[:2],
                preprocess_fn=self.preprocess_fn,
                stop_on_empty_dataset=False,
                include_features=self.params.mvm,
                reshuffle=False,
                resample=False,
                excluded_labels=self.excluded_labels,
                remapped_labels=self.remapped_labels,
                multi_label=self.params.multi_label,
                num_frames=self.params.square_width**2,
                channels=self.params.channels,
            )
            if self.test:
                test_accuracy = self.model.evaluate(self.test)

        self.save(run_name, history=history, test_results=test_accuracy)

    def checkpoints(self, run_name):
        checkpoint_file = self.checkpoint_folder / run_name / "cp.weights.h5"

        cp_callback = tf.keras.callbacks.ModelCheckpoint(
            filepath=checkpoint_file, save_weights_only=True, verbose=1
        )
        val_loss = self.checkpoint_folder / run_name / "val_loss.weights.h5"

        checkpoint_loss = tf.keras.callbacks.ModelCheckpoint(
            val_loss,
            monitor="val_loss",
            verbose=1,
            save_best_only=True,
            save_weights_only=True,
            mode="auto",
        )
        val_acc = self.checkpoint_folder / run_name / "val_acc.weights.h5"

        checkpoint_acc = tf.keras.callbacks.ModelCheckpoint(
            val_acc,
            monitor=(
                "val_binary_accuracy"
                if self.params.multi_label
                else "val_categorical_accuracy"
            ),
            verbose=1,
            save_best_only=True,
            save_weights_only=True,
            mode="max",
        )

        val_precision = self.checkpoint_folder / run_name / "val_recall.weights.h5"

        checkpoint_recall = tf.keras.callbacks.ModelCheckpoint(
            val_precision,
            monitor="val_recall",
            verbose=1,
            save_best_only=True,
            save_weights_only=True,
            mode="max",
        )
        earlyStopping = tf.keras.callbacks.EarlyStopping(
            patience=22,
            monitor=(
                "val_binary_accuracy"
                if self.params.multi_label
                else "val_categorical_accuracy"
            ),
            mode="max",
        )
        # havent found much use in this just takes training time
        # file_writer_cm = tf.summary.create_file_writer(
        #     self.log_base + "/{}/cm".format(run_name)
        # )
        # cm_callback = keras.callbacks.LambdaCallback(
        #     on_epoch_end=lambda epoch, logs: log_confusion_matrix(
        #         epoch, logs, self.model, self.test, file_writer_cm
        #     )
        # )
        #         "lr_callback": {
        #   "monitor": "val_categorical_accuracy",
        #   "mode": "max",
        #   "factor": 0.65,
        #   "patience": 15,
        #   "min_lr": 0.00002,
        #   "verbose": 1
        # },
        reduce_lr_callback = tf.keras.callbacks.ReduceLROnPlateau(
            monitor=(
                "val_binary_accuracy"
                if self.params.multi_label
                else "val_categorical_accuracy"
            ),
            mode="max",
            verbose=1,
        )
        return [
            earlyStopping,
            checkpoint_acc,
            checkpoint_loss,
            reduce_lr_callback,
            cp_callback,
        ]

    @property
    def hyperparams_string(self):
        """Returns list of hyperparameters as a string."""
        return "\n".join(
            ["{}={}".format(param, value) for param, value in self.params.items()]
        )

    def add_lstm(self, cnn):
        input_layer = tf.keras.Input(shape=(None, *self.params.output_dim))
        encoded_frames = tf.keras.layers.TimeDistributed(cnn)(input_layer)
        lstm_outputs = tf.keras.layers.LSTM(
            self.params["lstm_units"],
            dropout=self.params["keep_prob"],
            return_state=False,
        )(encoded_frames)

        hidden_layer = tf.keras.layers.Dense(1024, activation="relu")(lstm_outputs)
        hidden_layer = tf.keras.layers.Dense(512, activation="relu")(hidden_layer)

        preds = tf.keras.layers.Dense(
            len(self.labels), activation="softmax", name="pred"
        )(hidden_layer)
        model = tf.keras.models.Model(input_layer, preds)
        return model

    def classify_ir(self, clip, track, segment_frames=None):
        data = []
        thermal_median = np.empty(len(track.bounds_history), dtype=np.uint16)
        frames_used = []
        for i, region in enumerate(track.bounds_history):
            if region.blank:
                continue
            if region.width == 0 or region.height == 0:
                logging.warn(
                    "No width or height for frame %s regoin %s",
                    region.frame_number,
                    region,
                )
                continue
            frame = clip.frame_buffer.get_frame(region.frame_number)
            if frame is None:
                logging.error(
                    "Clasifying clip %s track %s can't get frame %s",
                    clip.get_id(),
                    track.get_id(),
                    region.frame_number,
                )
                raise Exception(
                    "Clasifying clip {} track {} can't get frame {}".format(
                        clip.get_id(), track.get_id(), region.frame_number
                    )
                )
            logging.debug(
                "classifying ir with preprocess %s size %s crop? %s f shape %s region %s",
                self.preprocess_fn.__module__,
                self.params.frame_size,
                crop,
                frame.thermal.shape,
                region,
            )
            preprocessed = preprocess_ir(
                frame,
                (
                    self.params.frame_size,
                    self.params.frame_size,
                ),
                region,
                self.preprocess_fn,
                save_info=f"{region.frame_number} - {region}",
            )
            logging.debug(
                "preprocessed is %s max %s min %s",
                preprocessed.shape,
                np.amax(preprocessed),
                np.amin(preprocessed),
            )
            frames_used.append(region.frame_number)
            data.append(preprocessed)
        if len(data) == 0:
            return None
        data = np.float32(data)
        track_prediction = TrackPrediction(track.get_id(), self.labels)

        output = self.model.predict(data)
        track_prediction.classified_clip(output, output, np.array(frames_used))
        track_prediction.normalize_score()
        return track_prediction

    def predict(self, frames):
        return self.model.predict(frames)

    def confusion_tracks(self, dataset, filename, threshold=0.8):
        logging.info("Calculating confusion with threshold %s", threshold)
        true_categories = []
        track_ids = []
        avg_mass = []
        for x, y in dataset:
            true_categories.extend(y[0].numpy())
            # dataset_y[0]
            track_ids.extend(y[1].numpy())
            avg_mass.extend(y[2].numpy())
        if len(true_categories) > 1:
            if self.params.multi_label:
                # multi = []
                # for y in true_categories:
                # multi.append(tf.where(y).numpy().ravel())
                # print(y, tf.where(y))
                # true_categories = np.int64(true_categories)
                pass
            else:
                true_categories = np.int64(tf.argmax(true_categories, axis=1))
        y_pred = self.model.predict(dataset)
        pred_per_track = {}

        # if self.params.multi_label:
        self.labels.append("nothing")
        # predicted_categori/es = []
        # for p in y_pred:
        # predicted_categories.append(tf.where(p >= 0.8).numpy().ravel())
        # predicted_categories = np.int64(predicted_categories)

        flat_y = []
        for y, track_id, mass, p in zip(true_categories, track_ids, avg_mass, y_pred):
            if self.params.multi_label:
                y_max = np.argmax(y)
            else:
                y_max = y
            track_pred = pred_per_track.setdefault(
                track_id, (y_max, TrackPrediction(track_id, self.labels))
            )
            track_pred[1].classified_frame(None, p, mass)

        results = [
            ("No-Smoothing", []),
            ("Old-Smoothing", []),
            ("New-Smoothing", []),
        ]
        for y, pred in pred_per_track.values():
            pred.normalize_score()
            preds = np.array([p.prediction for p in pred.predictions])

            no_smoothing = np.mean(preds, axis=0)
            masses = np.array(pred.masses)[:, None]
            old_smoothing = pred.class_best_score
            new_smooth = preds * masses
            new_smooth = np.sum(new_smooth, axis=0)
            new_smooth /= np.sum(masses)

            for i, pred_type in enumerate([no_smoothing, old_smoothing, new_smooth]):
                best_pred = np.argmax(pred_type)
                confidence = pred_type[best_pred]
                if confidence < threshold:
                    best_pred = len(self.labels) - 1  # Nothing
                results[i][1].append(best_pred)
            flat_y.append(y)

        true_categories = np.int64(flat_y)
        # else:
        #     predicted_categories = np.int64(tf.argmax(y_pred, axis=1))
        for result in results:
            cm = confusion_matrix(
                true_categories, np.int64(result[1]), labels=np.arange(len(self.labels))
            )
            # Log the confusion matrix as an image summary.
            figure = plot_confusion_matrix(cm, class_names=self.labels)
            plt.savefig(f"{result[0]}-{filename}", format="png")

    def confusion_tfrecords(self, dataset, filename):
        true_categories = tf.concat([y for x, y in dataset], axis=0)
        if len(true_categories) > 1:
            if self.params.multi_label:
                # multi = []
                # for y in true_categories:
                # multi.append(tf.where(y).numpy().ravel())
                # print(y, tf.where(y))
                # true_categories = np.int64(true_categories)
                pass
            else:
                true_categories = np.int64(tf.argmax(true_categories, axis=1))
        y_pred = self.model.predict(dataset)
        if self.params.multi_label:
            self.labels.append("nothing")
            # predicted_categori/es = []
            # for p in y_pred:
            # predicted_categories.append(tf.where(p >= 0.8).numpy().ravel())
            # predicted_categories = np.int64(predicted_categories)

            flat_p = []
            flat_y = []
            for y, p in zip(true_categories, y_pred):
                index = 0
                for y_l, p_l in zip(y, p):
                    predicted = p_l >= 0.8
                    if y_l == 0 and predicted:
                        flat_y.append(len(self.labels) - 1)
                        flat_p.append(index)
                    elif y_l == 1 and predicted:
                        flat_y.append(index)
                        flat_p.append(index)
                    elif y_l == 1 and not predicted:
                        flat_y.append(index)
                        flat_p.append(len(self.labels) - 1)
                    # elif y_l == 0 and not predicted:
                    # all good
                    # continue
                    index += 1
            true_categories = np.int64(flat_p)
            predicted_categories = np.int64(flat_y)
        else:
            predicted_categories = np.int64(tf.argmax(y_pred, axis=1))

        cm = confusion_matrix(
            true_categories, predicted_categories, labels=np.arange(len(self.labels))
        )
        np.save(str(Path(filename).with_suffix(".npy")), cm)
        # Log the confusion matrix as an image summary.
        figure = plot_confusion_matrix(cm, class_names=self.labels)
        plt.savefig(filename, format="png")

    # Obsolete
    def f1(self, batch_y, pred_raw):
        import tensorflow_addons as tfa

        one_hot_y = tf.keras.utils.to_categorical(batch_y, num_classes=len(self.labels))
        metric = tfa.metrics.F1Score(num_classes=len(self.labels))
        metric.update_state(one_hot_y, pred_raw)
        result = metric.result().numpy()
        logging.info("F1 score")
        by_label = {}
        for i, label in enumerate(self.labels):
            by_label[label] = round(100 * result[i])
        sorted = self.labels.copy()
        sorted.sort()
        for label in sorted:
            logging.info("%s = %s", label, by_label[label])

    def evaluate(self, dataset):
        test_accuracy = self.model.evaluate(dataset)
        logging.info("Test accuracy is %s", test_accuracy)

    #
    # # needs to be updated to work with tfrecord datagen
    # def track_accuracy(self, dataset, confusion="confusion.png"):
    #     dataset.set_read_only(True)
    #     dataset.use_segments = self.params.use_segments
    #     predictions = []
    #     actual = []
    #     raw_predictions = []
    #     total = 0
    #     correct = 0
    #     samples_by_label = {}
    #     incorrect_labels = {}
    #     for sample in dataset.segments:
    #         label_samples = samples_by_label.setdefault(sample.label, {})
    #         if sample.track_id in label_samples:
    #             label_samples[sample.track_id].append(sample)
    #         else:
    #             label_samples[sample.track_id] = [sample]
    #     bird_tracks = len(label_samples.get("bird", []))
    #
    #     for label in dataset.label_mapping.keys():
    #         incorrect = {}
    #         incorrect_labels[label] = incorrect
    #         track_samples = samples_by_label.get(label)
    #         if not track_samples:
    #             logging.warn("No samples for %s", label)
    #             continue
    #         track_samples = track_samples.values()
    #         if label == "insect" or label == "false-positive":
    #             track_samples = np.random.choice(
    #                 list(track_samples),
    #                 min(len(track_samples), bird_tracks),
    #                 replace=False,
    #             )
    #         logging.info("taking %s tracks for %s", len(track_samples), label)
    #         mapped_label = dataset.mapped_label(label)
    #         for track_segments in track_samples:
    #             segment_db = dataset.numpy_data.load_segments(track_segments)
    #             frame_db = {}
    #             for frames in segment_db.values():
    #                 for f in frames:
    #                     frame_db[f.frame_number] = f
    #             track_prediction = self.classify_track_data(
    #                 track_segments[0].track_id,
    #                 frame_db,
    #                 segments=track_segments,
    #             )
    #
    #             total += 1
    #             if track_prediction is None or len(track_prediction.predictions) == 0:
    #                 logging.warn("No predictions for %s", track_segments[0].track_id)
    #                 continue
    #             avg = np.mean(track_prediction.predictions, axis=0)
    #             actual.append(self.labels.index(mapped_label))
    #             predictions.append(track_prediction.best_label_index)
    #
    #             raw_predictions.append(avg)
    #             if actual[-1] == predictions[-1]:
    #                 correct += 1
    #             else:
    #                 if track_prediction.predicted_tag() in incorrect:
    #                     incorrect[track_prediction.predicted_tag()].append(
    #                         track_segments[0].unique_track_id
    #                     )
    #                 else:
    #                     incorrect[track_prediction.predicted_tag()] = [
    #                         track_segments[0].unique_track_id
    #                     ]
    #
    #             if total % 50 == 0:
    #                 logging.info("Processed %s", total)
    #     for label, incorrect in incorrect_labels.items():
    #         logging.info("Incorrect ************ %s", label)
    #         logging.info(incorrect.get("false-positive"))
    #     logging.info("Predicted correctly %s", round(100 * correct / total))
    #     self.f1(actual, raw_predictions)
    #
    #     if confusion is not None:
    #         cm = confusion_matrix(
    #             actual, predictions, labels=np.arange(len(self.labels))
    #         )
    #         figure = plot_confusion_matrix(cm, class_names=self.labels)
    #         plt.savefig(confusion, format="png")


# from tensorflow examples
def plot_confusion_matrix(cm, class_names):
    """
    Returns a matplotlib figure containing the plotted confusion matrix.

    Args:
      cm (array, shape = [n, n]): a confusion matrix of integer classes
      class_names (array, shape = [n]): String names of the integer classes
    """

    figure = plt.figure(figsize=(8, 8))
    plt.imshow(cm, interpolation="nearest", cmap=plt.cm.Blues)
    plt.title("Confusion matrix")
    plt.colorbar()
    tick_marks = np.arange(len(class_names))
    plt.xticks(tick_marks, class_names, rotation=45)
    plt.yticks(tick_marks, class_names)

    # Use white text if squares are dark; otherwise black.
    counts = cm.copy()
    threshold = counts.max() / 2.0

    # Normalize the confusion matrix.

    cm = np.around(cm.astype("float") / cm.sum(axis=1)[:, np.newaxis], decimals=2)
    cm = np.nan_to_num(cm)
    cm = np.uint8(np.round(cm * 100))

    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        color = "white" if counts[i, j] > threshold else "black"
        plt.text(j, i, cm[i, j], horizontalalignment="center", color=color)

    plt.tight_layout()
    plt.ylabel("True label")
    plt.xlabel("Predicted label")
    return figure


def log_confusion_matrix(epoch, logs, model, dataset, writer):
    # Use the model to predict the values from the validation dataset.

    true_categories = tf.concat([y for x, y in dataset], axis=0)
    true_categories = np.int64(tf.argmax(true_categories, axis=1))
    y_pred = model.model.predict(dataset)

    predicted_categories = np.int64(tf.argmax(y_pred, axis=1))

    cm = confusion_matrix(
        true_categories, predicted_categories, labels=np.arange(len(model.labels))
    )
    # Log the confusion matrix as an image summary.
    figure = plot_confusion_matrix(cm, class_names=model.labels)
    cm_image = plot_to_image(figure)

    # Log the confusion matrix as an image summary.
    with writer.as_default():
        tf.summary.image("Confusion Matrix", cm_image, step=epoch)


def plot_to_image(figure):
    """Converts the matplotlib plot specified by 'figure' to a PNG image and
    returns it. The supplied figure is closed and inaccessible after this call."""
    # Save the plot to a PNG in memory.
    buf = io.BytesIO()
    plt.savefig(buf, format="png")
    # Closing the figure prevents it from being displayed directly inside
    # the notebook.
    plt.close(figure)
    buf.seek(0)
    # Convert PNG buffer to TF image
    image = tf.image.decode_png(buf.getvalue(), channels=4)
    # Add the batch dimension
    image = tf.expand_dims(image, 0)
    return image


def loss(params):
    if params.multi_label:
        return tf.keras.losses.BinaryCrossentropy(
            label_smoothing=params.label_smoothing,
        )
    return tf.keras.losses.CategoricalCrossentropy(
        label_smoothing=params.label_smoothing,
    )


def optimizer(params):
    if params.learning_rate_decay is not None:
        learning_rate = tf.keras.optimizers.schedules.ExponentialDecay(
            params.learning_rate,
            decay_steps=100000,
            decay_rate=params.learning_rate_decay,
            staircase=True,
        )
    else:
        learning_rate = params.learning_rate  # setup optimizer
    if learning_rate:
        optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)
    else:
        optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)
    return optimizer


def validate_model(model_file):
    return Path(model_file).exists()


# HYPER PARAM TRAINING OF A MODEL
#
HP_DENSE_SIZES = hp.HParam("dense_sizes", hp.Discrete([""]))
HP_MVM = hp.HParam("mvm", hp.Discrete([3.0]))

HP_BATCH_SIZE = hp.HParam("batch_size", hp.Discrete([64]))
HP_OPTIMIZER = hp.HParam("optimizer", hp.Discrete(["adam"]))
HP_LEARNING_RATE = hp.HParam("learning_rate", hp.Discrete([0.001]))
HP_EPSILON = hp.HParam("epislon", hp.Discrete([1e-7]))  # 1.0 and 0.1 for inception
HP_DROPOUT = hp.HParam("dropout", hp.Discrete([0.0]))
HP_RETRAIN = hp.HParam("retrain_layer", hp.Discrete([-1]))
HP_LEARNING_RATE_DECAY = hp.HParam("learning_rate_decay", hp.Discrete([1.0]))

METRIC_ACCURACY = "accuracy"
METRIC_LOSS = "loss"


# GRID SEARCH
def train_test_model(model, hparams, log_dir, writer, epochs):
    # if not self.model:
    train_files = model.data_dir + "/train"
    validate_files = model.data_dir + "/validation"
    test_files = model.data_dir + "/test"
    mvm = hparams[HP_MVM]
    if mvm >= 1.0:
        mvm = True
    else:
        mvm = False
    train, remapped = get_dataset(
        train_files,
        model.type,
        model.labels,
        batch_size=model.params.batch_size,
        image_size=model.params.output_dim[:2],
        preprocess_fn=model.preprocess_fn,
        include_features=mvm,
        augment=True,
        scale_epoch=4,
    )
    validate, remapped = get_dataset(
        validate_files,
        model.type,
        model.labels,
        batch_size=model.params.batch_size,
        image_size=model.params.output_dim[:2],
        preprocess_fn=model.preprocess_fn,
        include_features=mvm,
        augment=False,
        scale_epoch=4,
    )

    test, _ = get_dataset(
        test_files,
        model.type,
        model.labels,
        batch_size=model.params.batch_size,
        image_size=model.params.output_dim[:2],
        preprocess_fn=model.preprocess_fn,
        include_features=mvm,
        deterministic=True,
        augment=False,
    )

    labels = model.labels

    opt = None
    learning_rate = hparams[HP_LEARNING_RATE]
    epsilon = hparams[HP_EPSILON]

    if hparams[HP_OPTIMIZER] == "adam":
        opt = tf.keras.optimizers.Adam(learning_rate=learning_rate, epsilon=epsilon)
    else:
        opt = tf.keras.optimizers.SGD(learning_rate=learning_rate, epsilon=epsilon)
    model.model.compile(
        optimizer=opt, loss=loss(model.params), metrics=["accuracy"], run_eagerly=True
    )
    cm_callback = tf.keras.callbacks.LambdaCallback(
        on_epoch_end=lambda epoch, logs: log_confusion_matrix(
            epoch, logs, model, test, writer
        )
    )
    history = model.model.fit(
        train,
        validation_data=validate,
        epochs=epochs,
        shuffle=False,
        # callbacks=[cm_callback],
        verbose=2,
    )
    tf.keras.backend.clear_session()
    gc.collect()
    del model
    del train
    del validate
    gc.collect()
    return history


def grid_search(keras_model, epochs=1):
    dir_name = keras_model.log_dir + "/hparam_tuning"
    with tf.summary.create_file_writer(dir_name).as_default():
        hp.hparams_config(
            hparams=[
                HP_MVM,
                HP_BATCH_SIZE,
                HP_DENSE_SIZES,
                HP_LEARNING_RATE,
                HP_OPTIMIZER,
                HP_EPSILON,
                HP_RETRAIN,
                HP_DROPOUT,
                HP_LEARNING_RATE_DECAY,
            ],
            metrics=[
                hp.Metric(METRIC_ACCURACY, display_name="Accuracy"),
                hp.Metric(METRIC_LOSS, display_name="Loss"),
            ],
        )
    session_num = 0
    hparams = {}
    for mvm in HP_MVM.domain.values:
        for batch_size in HP_BATCH_SIZE.domain.values:
            for dense_size in HP_DENSE_SIZES.domain.values:
                for retrain_layer in HP_RETRAIN.domain.values:
                    for learning_rate in HP_LEARNING_RATE.domain.values:
                        for optimizer in HP_OPTIMIZER.domain.values:
                            for epsilon in HP_EPSILON.domain.values:
                                for dropout in HP_DROPOUT.domain.values:
                                    for (
                                        learning_rate_decay
                                    ) in HP_LEARNING_RATE_DECAY.domain.values:
                                        hparams = {
                                            HP_MVM: mvm,
                                            HP_DENSE_SIZES: dense_size,
                                            HP_BATCH_SIZE: batch_size,
                                            HP_LEARNING_RATE: learning_rate,
                                            HP_OPTIMIZER: optimizer,
                                            HP_EPSILON: epsilon,
                                            HP_RETRAIN: retrain_layer,
                                            HP_DROPOUT: dropout,
                                            HP_LEARNING_RATE_DECAY: learning_rate_decay,
                                        }

                                        dense_layers = []
                                        if dense_size != "":
                                            for i, size in enumerate(dense_size):
                                                dense_layers[i] = int(size)

                                        # for some reason cant have None values in hyper params array
                                        if learning_rate_decay == 1.0:
                                            learning_rate_decay = None
                                        keras_model.params["learning_rate_decay"] = (
                                            learning_rate_decay
                                        )
                                        if mvm >= 1.0:
                                            keras_model.params["mvm"] = True
                                            keras_model.params["hq_mvm"] = mvm > 1
                                            keras_model.params["forest"] = mvm > 2

                                        else:
                                            keras_model.params["mvm"] = False

                                        keras_model.build_model(
                                            dense_sizes=dense_layers,
                                            retrain_from=(
                                                None
                                                if retrain_layer == -1
                                                else retrain_layer
                                            ),
                                            dropout=None if dropout == 0.0 else dropout,
                                        )
                                        keras_model.model.summary()

                                        run_name = "run-%d" % session_num
                                        print("--- Starting trial: %s" % run_name)
                                        print({h.name: hparams[h] for h in hparams})
                                        run(
                                            keras_model,
                                            dir_name + "/" + run_name,
                                            hparams,
                                            epochs,
                                        )
                                        session_num += 1


def run(keras_model, log_dir, hparams, epochs):
    with tf.summary.create_file_writer(log_dir).as_default() as w:
        hp.hparams(hparams)  # record the values used in this trial
        history = train_test_model(keras_model, hparams, log_dir, w, epochs=epochs)
        val_accuracy = history.history["val_accuracy"]
        val_loss = history.history["val_loss"]
        # log_confusion_matrix(epochs, None, self.model, self.validate, None)

        for step, accuracy in enumerate(val_accuracy):
            loss = val_loss[step]
            tf.summary.scalar(METRIC_ACCURACY, accuracy, step=step)
            tf.summary.scalar(METRIC_LOSS, loss, step=step)


from tensorflow.keras.callbacks import Callback


class ClearMemory(Callback):
    def on_epoch_end(self, epoch, logs=None):
        print("epoch edned", epoch)
        gc.collect()
        tf.keras.backend.clear_session()


def get_excluded(type, multi_label=False):
    if type == "thermal":
        return thermaldataset.get_excluded(), thermaldataset.get_remapped(multi_label)
    else:
        return irdataset.get_excluded(), irdataset.get_remapped()


def get_dataset(
    pattern,
    type,
    labels,
    **args,
):
    if type == "thermal":
        return get_tf(thermaldataset.load_dataset, pattern, labels, **args)
    else:
        return get_tf(irdataset.load_dataset, pattern, labels, **args)


class MetaJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, SegmentType) or isinstance(obj, FrameTypes):
            return obj.name
        return json.JSONEncoder.default(self, obj)



from datetime import timedelta
from .rawframe import RawFrame
from piclassifier.telemetry import Telemetry


class IRFrame(RawFrame):
    def __init__(self, headers):
        super().__init__(headers)

    def get_telemetry_size(self):
        return 0

    def parse_telemetry(self, raw_bytes):
        t = Telemetry()
        return t



import logging
import os

from .train import train_model


# this is a good list for a full search, but will take a long time to run (days)
FULL_SEARCH_PARAMS = {
    "batch_size": [1, 2, 4, 8, 16, 32, 64],
    "learning_rate": [1e-3, 3e-4, 1e-4, 3e-5, 1e-5],
    "l2_reg": [0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8],
    "label_smoothing": [0, 0.01, 0.05, 0.1, 0.2],
    "keep_prob": [0, 0.1, 0.2, 0.5, 0.8, 1.0],
    "batch_norm": [True, False],
    "lstm_units": [64, 128, 256, 512],
    "enable_flow": [True, False],
    "augmentation": [True, False],
    "thermal_threshold": [-100, -20, -10, 0, 10, 20, 100],
    # these runs will be identical in their parameters, it gives a feel for the varience during training.
    "identical": [1, 2, 3, 4, 5],
}

# this checks just the important parameters, and only around areas that are likely to work well.
# I've also excluded the default values as these do not need to be tested again.
SHORT_SEARCH_PARAMS = {
    "batch_size": [8, 32],
    "l2_reg": [1e-2, 1e-3, 1e-4],
    "label_smoothing": [0, 0.05, 0.2],
    "keep_prob": [0.1, 0.4, 0.6, 1.0],
    "batch_norm": [False],
    "lstm_units": [128, 512],
    "enable_flow": [False],
    "augmentation": [False],
    "thermal_threshold": [-100, -20, -10, 0, 20, 100],
}


def axis_search(conf):
    """
    Evaluate each hyper-parameter individually against a reference.

    The idea here is to assess each parameter individually while holding all other parameters at their default.
    For optimal results this will need to be done multiple times, each time updating the defaults to their optimal
    values.
    """
    logging.info("Performing hyper parameter search.")

    results_filename = os.path.join(conf.train.train_dir, "search-results.txt")
    logging.info(f"Writing hyper-parameter search results to {results_filename}")
    tracker = JobTracker(results_filename)

    # run the reference job with default params
    run_job(tracker, "reference", conf)

    for param_name, param_values in SHORT_SEARCH_PARAMS.items():
        for param_value in param_values:
            job_name = f"{param_name}={param_value}"
            run_job(tracker, job_name, conf, {param_name: param_value})


def run_job(tracker, job_name, conf, hyper_params=None):
    """Run a job with given hyper parameters, and log its results."""
    if tracker.is_done(job_name):
        return
    if not hyper_params:
        hyper_params = {}

    print("-" * 60)
    print("Processing", job_name)
    print("-" * 60)

    model = train_model(job_name, conf, hyper_params)
    tracker.mark_done(job_name, model.eval_score, hyper_params)


class JobTracker:
    def __init__(self, filename):
        self.filename = filename
        open(self.filename, "w").close()  # Create/truncate

    def is_done(self, job_name):
        """Returns True if this job has been processed before."""
        with open(self.filename, "r") as f:
            return any(line.split(",", 1)[0] == job_name for line in f)

    def mark_done(self, job_name, score, hyper_params):
        with open(self.filename, "a") as f:
            f.write(
                "{}, {}: {}\n".format(
                    job_name,
                    score,
                    " ".join(f"{k}={v}" for (k, v) in hyper_params.items()),
                )
            )



import numpy as np
import cv2

min_left = 1
min_top = 1
max_right = 159
max_bottom = 119
edge_bounds = [min_left, min_top, max_right, max_bottom]


def preprocess_frame(frame, bound, out_dim):
    frame = frame.filtered
    if np.sum(frame) == 0:
        return None
    else:
        maxdim = max(frame.shape)
        if maxdim > out_dim:
            mindim = min(frame.shape)
            diff = maxdim - mindim
            if diff:
                f = np.zeros((maxdim, maxdim), dtype=np.float32)
                c0, c1 = center_position(
                    bound.left, bound.right, min_left, max_right, maxdim
                )
                r0, r1 = center_position(
                    bound.top, bound.bottom, min_top, max_bottom, maxdim
                )
                f[r0:r1, c0:c1] = frame
                frame = f
            if any(
                a == b
                for a, b in zip(
                    [bound.left, bound.top, bound.right, bound.bottom], edge_bounds
                )
            ):
                dim = out_dim
            else:
                dim = maxdim - int((maxdim - out_dim) / 4 + 1) * 4
            frame = cv2.resize(frame, (dim, dim), interpolation=cv2.INTER_AREA)
        else:
            if bound.left == min_left:
                frame = fill_cols(frame, True, out_dim)
            elif bound.right == max_right:
                frame = fill_cols(frame, False, out_dim)
            if bound.top == min_top:
                frame = fill_rows(frame, True, out_dim)
            elif bound.bottom == max_bottom:
                frame = fill_rows(frame, False, out_dim)

    frame = center_frame(frame, out_dim)
    frame = frame[:, :, np.newaxis]
    return frame


def fill_rows(frame, fill_high, space):
    f = np.zeros((space, frame.shape[1]))
    if fill_high:
        f[: frame.shape[0], :] = frame
    else:
        f[space - frame.shape[0] :, :] = frame
    return f


def fill_cols(frame, fill_high, space):
    f = np.zeros((frame.shape[0], space))
    if fill_high:
        f[:, : frame.shape[1]] = frame
    else:
        f[:, space - frame.shape[1] :] = frame
    return f


def center_frame(frame, out_dim):
    out_frame = np.zeros((out_dim, out_dim), np.float32)
    y_start = (out_dim - frame.shape[0]) // 2
    x_start = (out_dim - frame.shape[1]) // 2

    out_frame[
        y_start : y_start + frame.shape[0], x_start : x_start + frame.shape[1]
    ] = frame
    return out_frame


def center_position(low_bound, high_bound, low_limit, high_limit, space):
    size = high_bound - low_bound
    extra = space - size
    if extra > 0:
        if low_bound == low_limit:
            return 0, size
        elif high_bound == high_limit:
            return space - size, space
        else:
            pad = int(extra / 2) + np.random.randint(extra % 2 + 1)
            return pad, size + pad
    else:
        return 0, size


def sum_weighted(predicts, weights):
    return np.matmul(weights.T, predicts)



import logging
import numpy as np
from pathlib import Path
import joblib

from ml_tools.interpreter import Interpreter
from classify.trackprediction import TrackPrediction
import cv2
import time

FEAT_LABELS = [
    "sqrt_area",
    "elongation",
    "peak_snr",
    "mean_snr",
    "fill_factor",
    "move_1",
    "rel_move_1",
    "rel_x_move_1",
    "rel_y_move_1",
    "move_3",
    "rel_move_3",
    "rel_x_move_3",
    "rel_y_move_3",
    "move_5",
    "rel_move_5",
    "rel_x_move_5",
    "rel_y_move_5",
    "max_speed",
    "min_speed",
    "avg_speed",
    "max_speed_x",
    "min_speed_x",
    "avg_speed_x",
    "max_speed_y",
    "min_speed_y",
    "avg_speed_y",
    "max_rel_speed",
    "min_rel_speed",
    "avg_rel_speed",
    "max_rel_speed_x",
    "min_rel_speed_x",
    "avg_rel_speed_x",
    "max_rel_speed_y",
    "min_rel_speed_y",
    "avg_rel_speed_y",
    "hist_diff",
]

# EXTRA_FEATURES = [
#     "speed_distance_ratio",
#     "speed_ratio",
#     "burst_min",
#     "burst_max",
#     "birst_mean",
#     "burst_chance",
#     "burst_per_frame",
#     "total frames",
# ]

EXTRA = ["avg", "std", "max", "min", "diff"]

ALL_FEATURES = []
for extra_lbl in EXTRA:
    for f in FEAT_LABELS:
        ALL_FEATURES.append(f"{extra_lbl}-{f}")
# ALL_FEATURES.extend(EXTRA_FEATURES)

important_features = [
    "std-fill_factor",
    "max-peak_snr",
    "std-move_1",
    "max-fill_factor",
    "std-hist_diff",
    "diff-hist_diff",
    "max-hist_diff",
    "min-hist_diff",
    "diff-fill_factor",
    "max-sqrt_area",
    "std-mean_snr",
    "max-min_rel_speed",
    "min-fill_factor",
    "std-rel_move_1",
    "diff-rel_x_move_1",
    "diff-move_1",
    "std-sqrt_area",
    "avg-move_3",
    "diff-elongation",
    "diff-move_5",
    "std-min_speed_x",
    "max-max_speed_x",
    "avg-max_speed_y",
    "max-elongation",
    "diff-move_3",
    "max-rel_x_move_3",
]


def feature_mask(features_used):
    feature_indexes = []
    for f in features_used:
        feature_indexes.append(ALL_FEATURES.index(f))
    feature_indexes = np.array(feature_indexes)
    return feature_indexes


def normalize(features):
    from ml_tools.featurenorms import mean_v, std_v

    features -= mean_v
    features /= std_v
    return features


class ForestModel(Interpreter):
    TYPE = "RandomForest"

    def __init__(self, model_file, data_type=None):
        super().__init__(model_file)
        model_file = Path(model_file)
        self.model = joblib.load(model_file)
        self.buffer_length = self.params.get("buffer_length", 1)
        self.features_used = self.params.get("features_used")
        self.features = self.params.get("features")
        self.mgrid = np.mgrid[:120, :160]
        # sligtly faster to reuse this mgrd

    def classify_track(
        self, clip, track, last_x_frames=None, segment_frames=None, min_segments=None
    ):

        track_prediction = TrackPrediction(track.get_id(), self.labels)
        predictions, frames, masses = self.predict_track(
            clip, track, last_x_frames=last_x_frames, normalize=True
        )
        track_prediction.classified_clip(predictions, predictions, frames, masses)
        return track_prediction

    def shape(self):
        return 1, (1, len(self.features))

    def preprocess(self):
        return

    def predict(self, x):
        return self.model.predict_proba(x)

    def predict_track(self, clip, track, **args):
        predict_from_last = args.get(
            "predict_from_last",
        )
        max_frames = args.get(
            "max_frames",
        )
        scale = args.get("scale")
        result = process_track(
            clip,
            track,
            self.mgrid,
            predict_from_last=predict_from_last,
            max_frames=max_frames,
            scale=scale,
            normalize=args.get("normalize", True),
            buf_len=self.buffer_length,
            last_frame_predicted=args.get("last_frame_predicted"),
        )
        if result is None:
            return None
        x, frames, masses = result
        if self.features_used is not None and len(self.features_used) > 0:
            f_mask = feature_mask(self.features_used)
            x = np.take(x, f_mask)

        # x = x[np.newaxis, :]
        predictions = self.model.predict_proba(x)
        # print("predictions", predictions.shape)
        return frames, predictions, masses


def process_track(
    clip,
    track,
    mgrid,
    predict_from_last=None,
    max_frames=None,
    buf_len=5,
    scale=None,
    normalize=True,
    last_frame_predicted=None,
):
    background = clip.background
    all_frames = None
    frame_temp_median = {}
    available_frames = len(clip.frame_buffer)
    if predict_from_last is None:
        bounds = track.bounds_history
        if last_frame_predicted is not None:
            last_track_frame = bounds[-1].frame_number
            bounds = bounds[-(last_track_frame - last_frame_predicted) :]

        if len(bounds) == 0:
            return None
        first_frame = bounds[0].frame_number
        last_frame = bounds[-1].frame_number
    else:
        bounds = track.bounds_history[-min(available_frames, predict_from_last) :]
        if last_frame_predicted is not None:
            last_track_frame = bounds[-1].frame_number
            bounds = bounds[-(last_track_frame - last_frame_predicted) :]
        all_frames = clip.frame_buffer.get_last_x(len(bounds))
        if len(all_frames) == 0:
            return None

        # possibility of last bound in track not being the last frame
        first_frame = all_frames[0].frame_number
        last_frame = all_frames[-1].frame_number

    indices = [
        i
        for i, region in enumerate(bounds)
        if not region.blank
        and region.width > 0
        and region.height > 0
        and region.frame_number >= first_frame
        and region.frame_number <= last_frame
    ]
    if len(indices) == 0:
        logging.info("No valid regions to predict on  track %s", track)
        return None
    logging.debug(
        "taking %s from %s with scale %s", len(bounds), len(track.bounds_history), scale
    )
    frames = []
    if clip.crop_rectangle is None:
        logging.warning("Clip has no crop rectangle")

    # iterator =enumerate(range(len(bounds)))
    if max_frames is not None:
        if len(indices) > max_frames:
            indices = np.random.choice(indices, max_frames, replace=False)
            indices.sort()

    data_bounds = np.empty(len(indices), dtype="O")

    for i, frame_i in enumerate(indices):
        region = bounds[frame_i].copy()
        data_bounds[i] = region
        if clip.crop_rectangle is not None:
            region.crop(clip.crop_rectangle)

        if all_frames is None:
            frame = clip.get_frame(region.frame_number)
        else:
            frame_index = region.frame_number - last_frame - 1
            # logging.info("Frame index is %s frame_i is %s",frame_index, frame_i)
            frame = all_frames[frame_index]
        frames.append(frame)
        frame_temp_median[region.frame_number] = np.median(frame.thermal)
        assert frame.frame_number == region.frame_number
    if scale is not None and scale != 1:
        resize = 1 / scale
        background = clip.rescaled_background(
            (int(background.shape[1] * resize), int(background.shape[0] * resize))
        )
    else:
        background = clip.background
    start = time.time()
    x, frames_used, masses = forest_features(
        frames,
        background,
        frame_temp_median,
        data_bounds,
        mgrid,
        cropped=False,
        normalize=normalize,
        buf_len=buf_len,
    )
    # logging.info("Feature time was %s",time.time()-start)
    return x, frames_used, masses


def forest_features(
    track_frames,
    background,
    frame_temp_median,
    regions,
    mgrid,
    buf_len=1,
    cropped=True,
    normalize=True,
):
    frame_features = []
    avg_features = None
    maximum_features = None
    minimum_features = None
    all_features = []
    f_count = 0
    prev_count = 0
    frames_used = []
    masses = []
    back_med = np.median(background)
    if len(track_frames) < buf_len:
        return None, None, None
    for i, frame in enumerate(track_frames):
        region = regions[i]
        if region.blank or region.width <= 0 or region.height <= 0:
            prev_count = 0
            continue
        frames_used.append(region.frame_number)
        masses.append(region.mass)
        feature = FrameFeatures(region)
        sub_back = region.subimage(background).copy()
        t_median = frame_temp_median[frame.frame_number]
        if cropped:
            cropped_frame = frame
        else:
            cropped_frame = frame.crop_by_region(region)
        thermal = cropped_frame.thermal
        # feature.calc_histogram(sub_back, thermal, normalize=normalize)

        f_count += 1

        thermal = thermal + back_med - t_median

        feature.calculate(thermal, cropped_frame.filtered, sub_back, mgrid)
        if buf_len > 1:
            count_back = min(buf_len, prev_count)

            for i in range(count_back):
                prev = frame_features[-i - 1]
                vel = feature.cent - prev.cent
                feature.speed[i] = np.sqrt(np.sum(vel * vel))
                feature.rel_speed[i] = feature.speed[i] / feature.sqrt_area
                feature.rel_speed_x[i] = np.abs(vel[0]) / feature.sqrt_area
                feature.rel_speed_y[i] = np.abs(vel[1]) / feature.sqrt_area
                feature.speed_x[i] = np.abs(vel[0])
                feature.speed_y[i] = np.abs(vel[1])

            frame_features.append(feature)
        features = feature.features()
        all_features.append(features)
        prev_count += 1
        if buf_len > 1:
            if maximum_features is None:
                maximum_features = features.copy()
                minimum_features = features.copy()

                avg_features = features.copy()
            else:
                maximum_features = np.maximum(features, maximum_features)
                non_zero = features != 0
                current_zero = minimum_features == 0
                minimum_features[current_zero] = features[current_zero]
                minimum_features[non_zero] = np.minimum(
                    minimum_features[non_zero], features[non_zero]
                )
                # Aggregate
                avg_features += features
    if f_count < buf_len:
        logging.error("Count is less than buff len %s %s", f_count, buf_len)
        return None
    # Compute statistics for all tracks that have the min required duration
    if buf_len == 1:
        return np.array(all_features), frames_used, masses
    else:
        N = f_count - np.array(
            [
                0,
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                1,
                3,
                3,
                3,
                3,
                5,
                5,
                5,
                5,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
            ]
        )  # Normalise each measure by however many samples went into it
        avg_features /= N
        std_features = np.sqrt(np.sum((all_features - avg_features) ** 2, axis=0) / N)
        diff_features = maximum_features - minimum_features
        burst_features = calculate_burst_features(frame_features, avg_features[5])

        X = np.hstack(
            (
                avg_features,
                std_features,
                maximum_features,
                minimum_features,
                diff_features,
                burst_features,
                np.array([len(track_frames)]),
            )
        )

        return X, frames_used, masses


def calculate_burst_features(frames, mean_speed):
    #

    cut_off = max(2, (1 + mean_speed))
    speed_above = len([f for f in frames if f.speed[0] > cut_off])
    speed_below = len([f for f in frames if f.speed[0] <= cut_off])

    burst_frames = 0
    burst_ratio = []
    burst_history = []
    total_birst_frames = 0
    low_speed_distance = 0
    high_speed_distance = 0
    for i, frame in enumerate(frames):
        if frame.speed[0] < cut_off:
            low_speed_distance += frame.speed[0]
        else:
            high_speed_distance += frame.speed[0]
        if i > 0:
            prev = frames[i - 1]
            if prev.speed[0] > cut_off and frame.speed[0] > cut_off:
                burst_frames += 1
            else:
                if burst_frames > 0:
                    burst_start = i - burst_frames - 1
                    if len(burst_history) > 0:
                        # length of non burst frames is from previous burst end
                        prev = burst_history[-1]
                        burst_start -= prev[0] + prev[1]
                    burst_history.append((i - burst_frames - 1, burst_frames + 1))
                    burst_ratio.append(burst_start / (burst_frames + 1))
                    total_birst_frames += burst_frames + 1
                    burst_frames = 0
    burst_ratio = np.array(burst_ratio)
    if speed_above == 0:
        speed_ratio = 0
        speed_distance_ratio = 0
    else:
        speed_distance_ratio = low_speed_distance / high_speed_distance
        speed_ratio = speed_below / speed_above

    if len(burst_ratio) == 0:
        burst_min = 0
        burst_max = 0
        burst_mean = 0
    else:
        burst_min = np.amin(burst_ratio)
        burst_max = np.amax(burst_ratio)
        burst_mean = np.mean(burst_ratio)
    burst_chance = len(burst_ratio) / len(frames)
    burst_per_frame = total_birst_frames / len(frames)
    return np.array(
        [
            speed_distance_ratio,
            speed_ratio,
            burst_min,
            burst_max,
            burst_mean,
            burst_chance,
            burst_per_frame,
        ]
    )


class FrameFeatures:
    def __init__(self, region, buff_len=5):
        # self.thermal = thermal
        self.region = region
        self.cent = None
        self.extent = None
        self.thera = None

        self.sqrt_area = None
        self.std_back = None
        self.peak_snr = None
        self.mean_snr = None
        self.fill_factor = None
        self.histogram_diff = 0
        self.thermal_min = None

        self.thermal_max = None
        self.thermal_std = None
        self.filtered_max = None
        self.filtered_std = None
        self.filtered_min = None

        self.rel_speed = np.zeros(buff_len)
        self.rel_speed_x = np.zeros(buff_len)
        self.rel_speed_y = np.zeros(buff_len)
        self.speed_x = np.zeros(buff_len)
        self.speed_y = np.zeros(buff_len)
        self.speed = np.zeros(buff_len)
        self.histogram_diff = 0

    def calculate(self, thermal, filtered, sub_back, mgrid):
        self.thermal_min = np.min(thermal)
        self.thermal_max = np.amax(thermal)
        self.thermal_std = np.std(thermal)
        # filtered = thermal - sub_back
        filtered = np.abs(filtered)
        self.filtered_max = np.amax(filtered)
        self.filtered_min = np.amin(filtered)
        self.filtered_std = np.std(filtered)

        # Calculate weighted centroid and second moments etc
        cent, extent, theta = intensity_weighted_moments(filtered, mgrid, self.region)

        self.cent = cent
        self.extent = extent
        self.theta = theta
        # Instantaneous shape features
        area = np.pi * extent[0] * extent[1]
        self.sqrt_area = np.sqrt(area)
        self.elongation = extent[0] / extent[1]
        self.std_back = np.std(sub_back) + 1.0e-9

        # Instantaneous intensity features
        self.peak_snr = (self.thermal_max - np.mean(sub_back)) / self.std_back
        self.mean_snr = self.thermal_std / self.std_back
        self.fill_factor = np.sum(filtered) / area

    def features(self):
        return np.array(
            [
                self.sqrt_area,
                self.elongation,
                self.peak_snr,
                self.mean_snr,
                self.fill_factor,
                # self.histogram_diff,
                self.thermal_max,
                self.thermal_min,
                self.thermal_std,
                self.filtered_max,
                self.filtered_min,
                self.filtered_std,
            ]
        )
        # non_zero = np.array([s for s in self.speed if s > 0])
        # max_speed = 0
        # min_speed = 0
        # avg_speed = 0
        # if len(non_zero) > 0:
        #     max_speed = np.amax(non_zero)
        #     min_speed = np.amin(non_zero)
        #     avg_speed = np.mean(non_zero)

        # non_zero = np.array([s for s in self.speed_x if s > 0])
        # max_speed_x = 0
        # min_speed_x = 0
        # avg_speed_x = 0
        # if len(non_zero) > 0:
        #     max_speed_x = np.amax(non_zero)
        #     min_speed_x = np.amin(non_zero)
        #     avg_speed_x = np.mean(non_zero)

        # non_zero = np.array([s for s in self.speed_y if s > 0])
        # max_speed_y = 0
        # min_speed_y = 0
        # avg_speed_y = 0
        # if len(non_zero) > 0:
        #     max_speed_y = np.amax(non_zero)
        #     min_speed_y = np.amin(non_zero)
        #     avg_speed_y = np.mean(non_zero)

        # non_zero = np.array([s for s in self.rel_speed if s > 0])
        # max_rel_speed = 0
        # min_rel_speed = 0
        # avg_rel_speed = 0
        # if len(non_zero) > 0:
        #     max_rel_speed = np.amax(non_zero)
        #     min_rel_speed = np.amin(non_zero)
        #     avg_rel_speed = np.mean(non_zero)

        # non_zero = np.array([s for s in self.rel_speed_x if s > 0])
        # max_rel_speed_x = 0
        # min_rel_speed_x = 0
        # avg_rel_speed_x = 0
        # if len(non_zero) > 0:
        #     max_rel_speed_x = np.amax(non_zero)
        #     min_rel_speed_x = np.amin(non_zero)
        #     avg_rel_speed_x = np.mean(non_zero)

        # non_zero = np.array([s for s in self.rel_speed_y if s > 0])
        # max_rel_speed_y = 0
        # min_rel_speed_y = 0
        # avg_rel_speed_y = 0
        # if len(non_zero) > 0:
        #     max_rel_speed_y = np.amax(non_zero)
        #     min_rel_speed_y = np.amin(non_zero)
        #     avg_rel_speed_y = np.mean(non_zero)

        # return np.array(
        #     [
        #         self.sqrt_area,
        #         self.elongation,
        #         self.peak_snr,
        #         self.mean_snr,
        #         self.fill_factor,
        #         self.speed[0],
        #         self.rel_speed[0],
        #         self.rel_speed_x[0],
        #         self.rel_speed_y[0],
        #         self.speed[2],
        #         self.rel_speed[2],
        #         self.rel_speed_x[2],
        #         self.rel_speed_y[2],
        #         self.speed[4],
        #         self.rel_speed[4],
        #         self.rel_speed_x[4],
        #         self.rel_speed_y[4],
        #         max_speed,
        #         min_speed,
        #         avg_speed,
        #         max_speed_x,
        #         min_speed_x,
        #         avg_speed_x,
        #         max_speed_y,
        #         min_speed_y,
        #         avg_speed_y,
        #         max_rel_speed,
        #         min_rel_speed,
        #         avg_rel_speed,
        #         max_rel_speed_x,
        #         min_rel_speed_x,
        #         avg_rel_speed_x,
        #         max_rel_speed_y,
        #         min_rel_speed_y,
        #         avg_rel_speed_y,
        #         self.histogram_diff,
        #     ]
        # )

    def calc_histogram(self, sub_back, crop_t, normalize=False):
        if normalize:
            max_v = np.amax(sub_back)
            min_v = np.amin(sub_back)
            sub_back = (np.float32(sub_back) - min_v) / (max_v - min_v)
            max_v = np.amax(crop_t)
            min_v = np.amin(crop_t)
            crop_t = (np.float32(crop_t) - min_v) / (max_v - min_v)

            sub_back *= 255
            crop_t *= 255
        assert crop_t.shape == sub_back.shape
        # sub_back = np.uint8(sub_back)
        # crop_t = np.uint8(crop_t)
        sub_back = sub_back[..., np.newaxis]
        crop_t = crop_t[..., np.newaxis]
        h_bins = 60
        histSize = [h_bins]
        channels = [0]
        hist_base = cv2.calcHist(
            [sub_back],
            channels,
            None,
            histSize,
            [0, 255],
            accumulate=False,
        )
        cv2.normalize(hist_base, hist_base, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX)
        hist_track = cv2.calcHist(
            [crop_t],
            channels,
            None,
            histSize,
            [0, 255],
            accumulate=False,
        )
        cv2.normalize(
            hist_track,
            hist_track,
            alpha=0,
            beta=1,
            norm_type=cv2.NORM_MINMAX,
        )
        self.histogram_diff = cv2.compareHist(hist_base, hist_track, 0)


# Find centre of mass and size/orientation of the hot spot
def intensity_weighted_moments(sub, mgrid, region=None):
    tot = np.sum(sub)
    # print(tot, "using", region)
    if tot <= 0.0:
        # Zero image - replace with ones so calculations can continue
        sub = np.ones(sub.shape)
        tot = sub.size
        # surely all zeros in this case

    # Calculate weighted centroid
    Y = mgrid[0][: sub.shape[0], : sub.shape[1]]
    X = mgrid[1][: sub.shape[0], : sub.shape[1]]
    # Y, X = np.mgrid[0 : sub.shape[0], 0 : sub.shape[1]]

    cx = np.sum(sub * X) / tot
    cy = np.sum(sub * Y) / tot
    X = X - cx
    Y = Y - cy
    cent = np.array([region.x + cx, region.y + cy])

    # Second moments matrix
    mxx = np.sum(X * X * sub) / tot
    mxy = np.sum(X * Y * sub) / tot
    myy = np.sum(Y * Y * sub) / tot
    M = np.array([[mxx, mxy], [mxy, myy]])

    # Extent and angle
    w, v = np.linalg.eigh(M)
    w = np.abs(w)
    if w[0] < w[1]:
        w = w[::-1]
        v = v[:, ::-1]
    extent = (
        np.sqrt(w) + 0.5
    )  # Add half a pixel so that a single bright pixel has non-zero extent
    theta = np.arctan2(v[1, 0], v[0, 0])

    return cent, extent, theta



"""
Author: Matthew Aitchison
Date: December 2017

Dataset used for training a tensorflow model from track data.

Tracks are broken into segments.  Filtered, and then passed to the trainer using a weighted random sample.

"""

import logging
import math
import os
import time
import numpy as np
import gc
from ml_tools.datasetstructures import TrackHeader, TrackingSample, SegmentType
from ml_tools.trackdatabase import TrackDatabase
from ml_tools.rawdb import RawDatabase
from ml_tools import tools
from track.region import Region
import json
from config.buildconfig import BuildConfig
from pathlib import Path


class Dataset:
    """
    Stores visit, clip, track, and segment information headers in memory, and allows track / segment streaming from
    disk.
    """

    def __init__(
        self,
        dataset_dir,
        name="Dataset",
        config=None,
        use_predictions=False,
        consecutive_segments=False,
        labels=[],
        label_mapping=None,
        raw=True,
        ext=".cptv",
    ):
        self.ext = ext
        self.raw = raw
        # if self.raw and self.ext == ".cptv":
        # raise Exception("Raw CPTV not implemented yet")
        self.dataset_dir = Path(dataset_dir)
        self.consecutive_segments = consecutive_segments
        self.label_mapping = label_mapping
        # name of this dataset
        self.name = name
        # list of our tracks
        self.samples_by_bin = {}
        # self.samples = []
        self.samples_by_id = {}
        self.clips = []

        # list of label names
        self.labels = labels

        self.enable_augmentation = False
        self.label_caps = {}
        self.use_segments = True
        if config:
            self.tag_precedence = config.build.tag_precedence
            self.type = config.train.type
            if config.train.type == "IR":
                self.use_segments = False
                self.segment_length = 1
            else:
                self.use_segments = config.build.use_segments
                if self.use_segments:
                    self.segment_length = config.build.segment_length
                else:
                    self.segment_length = 1
            # number of seconds segments are spaced apart
            self.segment_spacing = config.build.segment_spacing
            self.banned_clips = config.build.banned_clips
            self.included_labels = config.labels
            self.segment_min_avg_mass = config.build.segment_min_avg_mass
            self.excluded_tags = config.build.excluded_tags
            self.min_frame_mass = config.build.min_frame_mass
            self.filter_by_lq = config.build.filter_by_lq
            self.segment_types = [SegmentType.ALL_RANDOM_MASKED]
            self.max_segments = config.build.max_segments
            self.country = config.build.country
            self.max_frames = config.build.max_frames
        else:
            self.country = "NZ"
            self.tag_precedence = BuildConfig.DEFAULT_GROUPS
            self.filter_by_lq = False
            # number of seconds each segment should be
            if self.use_segments:
                self.segment_length = 25
            else:
                self.segment_length = 1
            # number of seconds segments are spaced apart
            self.segment_spacing = 1
            self.segment_min_avg_mass = 10
            self.min_frame_mass = 16
            self.segment_types = [SegmentType.ALL_RANDOM]
            self.max_frames = 75

        self.country_rectangle = BuildConfig.COUNTRY_LOCATIONS.get(self.country)
        logging.info(
            "Filtering by country %s have boundying %s",
            self.country,
            self.country_rectangle,
        )
        self.max_frame_mass = None
        self.filtered_stats = {
            "confidence": 0,
            "trap": 0,
            "banned": 0,
            "date": 0,
            "tags": 0,
            "segment_mass": 0,
            "no_data": 0,
            "not-confirmed": 0,
            "tag_names": set(),
            "notags": 0,
            "bad_track_json": 0,
        }
        self.lbl_p = None
        self.numpy_data = None

        self.skip_ffc = True

    @property
    def samples(self):
        return self.samples_by_id.values()

    @property
    def sample_count(self):
        return len(self.samples_by_id)

    def get_label_caps(self, labels, remapped=False):
        counts = []
        for label in labels:
            counts.append(len(self.samples_for(label, remapped=remapped)))
        index = math.floor(len(counts) * 0.40)
        counts.sort()
        birds = self.samples_for("bird", remapped=remapped)
        if len(birds) > 0:
            return len(birds)

        return counts[index]

    def samples_for(self, label, remapped=False):
        labels = []
        if remapped and self.label_mapping:
            labels = [
                key
                for key, mapped in self.label_mapping.items()
                if mapped.lower() == label.lower()
            ]
            labels.sort()
        else:
            labels.append(label)
        samples = []
        for l in labels:
            samples.extend(self.samples_by_label.get(l, []))
        return samples

    def get_counts(self, label):
        """
        Gets number of examples for given label
        :label: label to check
        :return: (segments, tracks, bins, weight)
        """
        tracks = 0
        bins = 0
        weight = 0
        samples_count = 0
        samples = self.samples_by_label.get(label, [])
        tracks = len(set([sample.track_id for sample in samples]))
        weight = self.get_label_weight(label)
        bins = len(set([sample.bin_id for sample in samples]))
        samples_count = len(samples)
        return samples_count, tracks, bins, weight

    def load_clips(
        self,
        shuffle=False,
        before_date=None,
        after_date=None,
        label=None,
        dont_filter_segment=False,
    ):
        """
        Loads track headers from track database with optional filter
        :return: [number of tracks added, total tracks].
        """

        counter = 0
        logging.info("Loading clips")
        for db_clip in self.dataset_dir.glob(f"**/*{self.ext}"):
            tracks_added = self.load_clip(db_clip, dont_filter_segment)
            if tracks_added == 0:
                logging.info("No tracks added for %s", db_clip)
            counter += 1
            if counter % 50 == 0:
                logging.debug("Dataset loaded %s", counter)
        return [counter, counter]

    def load_clip(self, db_clip, dont_filter_segment=False):
        if self.raw:
            db = RawDatabase(db_clip)
        else:
            db = TrackDatabase(db_clip)
        try:
            clip_header = db.get_clip_tracks(self.tag_precedence)
        except:
            logging.error("Could not load %s", db_clip, exc_info=True)
            return 0
        if clip_header is None or filter_clip(
            clip_header,
            clip_header.location,
            self.country_rectangle,
            self.filtered_stats,
        ):
            return 0
        filtered = 0
        added = 0
        clip_header.tracks = [
            track
            for track in clip_header.tracks
            if not filter_track(track, self.excluded_tags, self.filtered_stats)
        ]
        self.clips.append(clip_header)
        for track_header in clip_header.tracks:
            if self.label_mapping:
                track_header.remapped_label = self.label_mapping.get(
                    track_header.original_label, track_header.original_label
                )
            added += 1
            if self.use_segments:
                segment_frame_spacing = int(
                    round(self.segment_spacing * clip_header.frames_per_second)
                )
                segment_width = self.segment_length
                track_header.get_segments(
                    segment_width,
                    segment_frame_spacing,
                    self.segment_types,
                    self.segment_min_avg_mass,
                    max_segments=self.max_segments,
                    dont_filter=dont_filter_segment,
                    skip_ffc=self.skip_ffc,
                    ffc_frames=clip_header.ffc_frames,
                )
                self.filtered_stats["segment_mass"] += track_header.filtered_stats[
                    "segment_mass"
                ]

            else:
                skip_last = None
                if self.type == "IR":
                    # a lot of ir clips have bad tracking near end so just reduce track length
                    skip_last = 0.1
                track_header.calculate_sample_frames(
                    min_mass=(
                        self.min_frame_mass
                        if not self.filter_by_lq
                        else track_header.lower_mass
                    ),
                    max_mass=(
                        self.max_frame_mass
                        if not self.filter_by_lq
                        else track_header.upper_mass
                    ),
                    ffc_frames=clip_header.ffc_frames,
                    skip_last=skip_last,
                )
                if track_header.label not in self.labels:
                    self.labels.append(track_header.label)
            for sample in track_header.samples:
                self.add_clip_sample_mappings(sample)
        return added

    def add_samples(self, samples):
        """
        Adds list of samples to dataset
        :param track_filter: optional filter
        """
        result = 0
        for sample in samples:
            if self.add_clip_sample_mappings(sample):
                result += 1
        return result

    @property
    def samples_by_label(self):
        samples_by_label = {}
        for sample in self.samples:
            if sample.label not in samples_by_label:
                samples_by_label[sample.label] = []
            samples_by_label[sample.label].append(sample)
        return samples_by_label

    def add_clip_sample_mappings(self, sample):
        self.samples_by_id[sample.id] = sample
        # self.samples[sample.id] = sample

        if self.label_mapping:
            sample.remapped_label = self.label_mapping.get(
                sample.original_label, sample.original_label
            )

        if sample.label not in self.labels:
            self.labels.append(sample.label)
        bins = self.samples_by_bin.setdefault(sample.bin_id, {})
        bins[sample.id] = sample
        # (sample)
        return True

    def epoch_samples(
        self, cap_samples=None, replace=True, random=True, cap_at=None, label_cap=None
    ):
        if len(self.labels) == 0:
            return []
        labels = self.labels.copy()
        samples = []

        if (cap_at or cap_samples) and label_cap is None:
            if cap_at:
                label_cap = len(self.samples_for(cap_at, remapped=True))
            else:
                label_cap = self.get_label_caps(labels, remapped=True)

        cap = None
        for label in labels:
            if label_cap:
                cap = min(label_cap, len(self.samples_for(label, remapped=True)))
            if label == "false-positive":
                if cap is None:
                    cap = int(label_cap * 0.5)
                else:
                    cap = min(cap, int(label_cap * 0.5))
            new = self.get_sample(cap=cap, replace=replace, label=label, random=random)
            if new is not None and len(new) > 0:
                samples.extend(new)
        labels = [sample.label for sample in samples]
        return samples

    def cdf(self):
        return self.sample_cdf

    def label_cdf(self, label):
        return self.sample_label_cdf.get(label, [])

    def get_samples_by_source(self):
        samples_by_source = {}
        for s in self.samples_by_id.values():
            samples_by_source.setdefault(s.source_file, []).append(s)
        return samples_by_source

    #
    # def get_sample(self, cap=None, replace=True, label=None, random=True):
    #     """Returns a random frames from weighted list."""
    #     if label:
    #         samples = self.samples_for(label, remapped=True)
    #         cdf = self.label_cdf(label)
    #     else:
    #         samples = self.samples()
    #         cdf = self.cdf()
    #     if not samples:
    #         return None
    #     if cap is None:
    #         return samples
    #     if random:
    #         return np.random.choice(samples, cap, replace=replace, p=cdf)
    #     else:
    #         cap = min(cap, len(samples))
    #         return samples[:cap]
    #
    # def balance_bins(self, max_bin_weight=None):
    #     """
    #     Adjusts weights so that bins with a number number of segments aren't sampled so frequently.
    #     :param max_bin_weight: bins with more weight than this number will be scaled back to this weight.
    #     """
    #
    #     for bin_name, samples in self.samples_by_bin.items():
    #         bin_weight = sum(sample.sample_weight for sample in samples)
    #         if bin_weight == 0:
    #             continue
    #         if max_bin_weight is None:
    #             scale_factor = 1 / bin_weight
    #             # means each bin has equal possiblity
    #         elif bin_weight > max_bin_weight:
    #             scale_factor = max_bin_weight / bin_weight
    #         else:
    #             scale_factor = 1
    #         for sample in samples:
    #             sample.weight = np.float16(sample.sample_weight * scale_factor)
    #     self.rebuild_cdf()

    def remove_label(self, label_to_remove):
        """
        Removes all segments of given label from dataset. Label remains in dataset.labels however, so as to not
        change the ordinal value of the labels.
        """
        if label_to_remove not in self.labels:
            return
        tracks = self.tracks_by_label[label_to_remove]
        for track in tracks:
            self.remove_label()

        self.rebuild_cdf()

    def mapped_label(self, label):
        if self.label_mapping:
            return self.label_mapping.get(label, label)
        return label

    #
    # def rebuild_cdf(self, lbl_p=None):
    #     """Calculates the CDF used for fast random sampling for frames and
    #     segments, if balance labels is set each label has an equal chance of
    #     being chosen
    #     """
    #     if lbl_p is None:
    #         lbl_p = self.lbl_p
    #
    #     self.sample_cdf = []
    #     total = 0
    #     self.sample_label_cdf = {}
    #
    #     for sample in self.samples:
    #         sample_weight = sample.sample_weight
    #         if lbl_p and sample.label in lbl_p:
    #             sample_weight *= lbl_p[sample.label]
    #         total += sample_weight
    #
    #         self.sample_cdf.append(sample_weight)
    #
    #         cdf = self.sample_label_cdf.setdefault(sample.label, [])
    #         cdf.append(sample.sample_weight)
    #
    #     if len(self.sample_cdf) > 0:
    #         self.sample_cdf = [x / total for x in self.sample_cdf]
    #
    #     for key, cdf in self.sample_label_cdf.items():
    #         total = sum(cdf)
    #         self.sample_label_cdf[key] = [x / total for x in cdf]
    #
    #     if self.label_mapping:
    #         mapped_cdf = {}
    #         labels = list(self.label_mapping.keys())
    #         labels.sort()
    #         for label in labels:
    #             if label not in self.sample_label_cdf:
    #                 continue
    #             label_cdf = self.sample_label_cdf[label]
    #             new_label = self.label_mapping[label]
    #             if lbl_p and label in lbl_p:
    #                 label_cdf = np.float64(label_cdf)
    #                 label_cdf *= lbl_p[label]
    #             cdf = mapped_cdf.setdefault(new_label, [])
    #             cdf.extend(label_cdf)
    #
    #         for key, cdf in mapped_cdf.items():
    #             total = sum(cdf)
    #             mapped_cdf[key] = [x / total for x in cdf]
    #         self.sample_label_cdf = mapped_cdf

    def get_label_weight(self, label):
        """Returns the total weight for all segments of given label."""
        samples = self.samples_by_label.get(label)
        return sum(sample.weight for sample in samples) if samples else 0

    def regroup(
        self,
        groups,
        shuffle=True,
    ):
        """
        regroups the dataset so multiple animals can be under a single label
        """
        self.label_mapping = groups
        samples_by_bin = {}
        samples_by_label = {}

        for lbl, remapped_lbl in groups.items():
            count = 0
            samples = self.samples_by_label[lbl]
            for s in samples:
                s.remapped_label = remapped_lbl
                samples_by_bin.setdefault(s.bin_id, {})[s.id] = s
                # .append(s)
            samples_by_label.setdefault(remapped_lbl, []).extend(samples)

        self.labels = list(groups.keys())
        self.labels.sort()
        self.samples_by_bin = samples_by_bin
        self.samples_by_label = samples_by_label
        #
        # if shuffle:
        #     np.random.shuffle(self.samples)
        # self.rebuild_cdf()

    def has_data(self):
        return len(self.samples_by_id) > 0

    def remove_sample_by_id(self, id, bin_id):
        del self.samples_by_id[id]
        try:
            # not nessesarily there if splitting by clip hack
            del self.samples_by_bin[bin_id][id]
        except:
            pass

    def remove_sample(self, sample):
        del self.samples_by_id[sample.id]
        try:
            # not nessesarily there if splitting by clip hack
            del self.samples_by_bin[sample.bin_id][sample.id]
        except:
            pass

    def fetch_track(
        self,
        track: TrackHeader,
        original=False,
        sample_frames=False,
    ):
        """
        Fetches data for an entire track
        :param track: the track to fetch
        :return: segment data of shape [frames, channels, height, width]
        """
        frame_numbers = None
        if sample_frames:
            frame_numbers = [frame.frame_num for frame in track.sample_frames]
            frame_numbers.sort()
        frames = self.db.get_track(
            track.clip_id,
            track.track_id,
            original=original,
            frame_numbers=frame_numbers,
        )
        return frames


def filter_track(track_header, excluded_tags, filtered_stats={}):
    # some clips are banned for various reasons4
    if track_header.label is None:
        filtered_stats.setdefault("notags", 0)
        filtered_stats["notags"] += 1
        return True

    if track_header.label in excluded_tags:
        filtered_stats.setdefault("tags", 0)

        filtered_stats["tags"] += 1
        filter_tags = filtered_stats.setdefault("tag_names", set())
        filter_tags.add(track_header.label)

        return True

    if track_header.human_tags is not None:
        found_tags = [
            tag[0] for tag in track_header.human_tags if tag[0] in excluded_tags
        ]
        if len(found_tags) > 0:
            filter_tags = filtered_stats.setdefault("tag_names", set())
            filter_tags |= set(found_tags)
            filtered_stats.setdefault("tags", 0)
            filtered_stats["tags"] += 1
            return True
    # always let the false-positives through as we need them even though they would normally
    # be filtered out.

    if track_header.regions_by_frame is None or len(track_header.regions_by_frame) == 0:
        filtered_stats.setdefault("no_data", 0)
        filtered_stats["no_data"] += 1
        logging.info("No region data")
        return True

    # dont think we need this gp 28/08/2023
    # if track_meta["human_tag"] == "false-positive":
    # return False

    # for some reason we get some records with a None confidence?
    if track_header.confidence <= 0.6:
        filtered_stats.setdefault("confidence", 0)
        filtered_stats["confidence"] += 1
        return True

    return False


def filter_clip(clip, location, location_bounds, filtered_stats=None, after_date=None):
    # remove tracks of trapped animals
    if (clip.events is not None and "trap" in clip.events.lower()) or (
        clip.trap is not None and "trap" in clip.trap.lower()
    ):
        if filtered_stats is not None:
            if "trap" in filtered_stats:
                filtered_stats["trap"] += 1
            else:
                filtered_stats["trap"] = 1
        logging.info("Filtered because in trap")
        return True

    if (
        location is not None
        and location_bounds is not None
        and not location_bounds.contains(*location)
    ):
        if filtered_stats is not None:
            if "location" in filtered_stats:
                filtered_stats["location"] += 1
            else:
                filtered_stats["location"] = 1
        return True

    if after_date is not None and clip.rec_time <= after_date:
        if filtered_stats is not None:
            if "date" in filtered_stats:
                filtered_stats["date"] += 1
            else:
                filtered_stats["date"] = 1
        return True

    return False



import argparse
import os
import numpy as np

import shutil
import tensorflow_decision_forests as tfdf

import tensorflow as tf
from config.config import Config
import pickle
from pathlib import Path
import shutil
from ml_tools.interpreter import LiteInterpreter

MODEL_DIR = "../cptv-download/train/checkpoints"
MODEL_NAME = "training-most-recent.sav"
SAVED_DIR = "saved_model"
LITE_MODEL_NAME = "converted_model.tflite"


def run_model(args):
    model = LiteInterpreter(args.model)
    input_data = np.array(np.random.random_sample(model.shape()[1:]), dtype=np.float32)
    prediction = model.predict(input_data)
    print("model pass 1 predicted", prediction)
    input_data = np.array(np.random.random_sample(model.shape()[1:]), dtype=np.float32)
    prediction = model.predict(input_data)
    print("model pass 2 predicted", prediction)


def convert_model(args):
    print("Loading: ", args.model)
    args.model = Path(args.model)
    model_dir = args.model.parent
    lite_dir = model_dir / "tflite"
    import time

    a = time.time()
    if args.model.suffix == ".pb":
        # for some reason refuses to work with absolute path
        model = tf.keras.models.load_model(args.model.parent, compile=False)
    else:
        model = tf.keras.models.load_model(args.model, compile=False)
    print(time.time() - a, " to load model")
    # return
    model.trainable = False
    meta_file = args.model.with_suffix(".json")

    if args.weights:
        print("using weights ", args.weights)
        model.load_weights(args.weights).expect_partial()
    if args.convert:
        converter = tf.lite.TFLiteConverter.from_keras_model(model)
        # converter.target_spec.supported_ops = [
        #     tf.lite.OpsSet.TFLITE_BUILTINS,  # enable TensorFlow Lite ops.
        #     tf.lite.OpsSet.SELECT_TF_OPS,  # enable TensorFlow ops.
        # ]
        # 8 bit ingeter
        converter.optimizations = [tf.lite.Optimize.DEFAULT]

        tflite_model = converter.convert()
        print("saving model to ", out_dir / args.model.stem)
        out_dir.mkdir(parents=True, exist_ok=True)
        with (out_dir / args.model.stem).open("wb") as f:
            f.write(tflite_model)
        frozen_meta = out_dir / meta_file.name

    elif args.freeze or args.export:
        out_dir = Path(args.freeze)
        out_dir.mkdir(parents=True, exist_ok=True)

        if args.export:
            model.summary()
            input_signature = get_input_sig(model)
            export_archive = tf.keras.export.ExportArchive()
            export_archive.track(model)
            export_archive.add_endpoint(
                name="predict", fn=model.call, input_signature=input_signature
            )
            export_archive.write_out(out_dir)

            print("saving model to", out_dir / "saved_model.pb")
            frozen_meta = out_dir / "saved_model.json"

        else:
            print("saving model to", out_dir / args.model.name)
            model.save(out_dir / args.model.name)
            frozen_meta = out_dir / meta_file.name

    if meta_file.exists():
        shutil.copy(meta_file, frozen_meta)


def get_input_sig(model):
    inputs = []
    for input in model.inputs:
        inputs.append(tf.TensorSpec(shape=input.shape, dtype=input.dtype))
    return inputs


def load_model(args):
    print("loading model ", args.model)
    model_dir = Path(args.model)
    if model_dir.is_file():
        model_dir = model_dir.parent
    lite_dir = model_dir / "tflite"
    print("loading", model_dir)
    model = tf.keras.models.load_model(str(model_dir))
    model.trainable = False
    if args.weights:
        print("using weights ", args.weights)
        model.load_weights(args.weights).expect_partial()
    return model


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-f",
        "--freeze",
        help="freeze model with weights here",
    )
    parser.add_argument(
        "-e",
        "--export",
        action="store_true",
        help="export model instead of saving",
    )
    parser.add_argument("-w", "--weights", help="Weights to use")

    parser.add_argument(
        "-c", "--convert", action="store_true", help="Convert frozen model to tflite"
    )
    parser.add_argument(
        "-r",
        "--run",
        action="store_true",
        help="Test converted model with random data using tflite interpreter",
    )
    parser.add_argument(
        "-m",
        "--model",
        default=MODEL_DIR,
        help="Directory where meta data of the model you want to convert is stored",
    )

    return parser.parse_args()


def main():
    args = parse_args()
    if args.run:
        run_model(args)
    else:
        convert_model(args)


if __name__ == "__main__":
    main()



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import ml_tools.tools as tools
from ml_tools.rectangle import Rectangle
import attr
import logging
import numpy as np


@attr.s(eq=False, slots=True)
class Region(Rectangle):
    """Region is a rectangle extended to support mass."""

    centroid = attr.ib()
    mass = attr.ib(default=0)
    # how much pixels in this region have changed since last frame
    frame_number = attr.ib(default=0)
    pixel_variance = attr.ib(default=0)
    id = attr.ib(default=0)

    # if this region was cropped or not
    was_cropped = attr.ib(default=False)
    blank = attr.ib(default=False)
    is_along_border = attr.ib(default=False)
    in_trap = attr.ib(default=False)

    def rescale(self, factor):
        self.x = int(self.x * factor)
        self.y = int(self.y * factor)

        self.width = int(self.width * factor)
        self.height = int(self.height * factor)
        self.mass = self.mass * (factor**2)

    @staticmethod
    def from_ltwh(left, top, width, height):
        """Construct a rectangle from left, top, right, bottom co-ords."""
        return Region(left, top, width=width, height=height, centroid=None)

    def to_array(self):
        """Return rectangle as left, top, right, bottom co-ords."""
        return np.uint16(
            [
                self.left,
                self.top,
                self.right,
                self.bottom,
                self.frame_number,
                self.mass,
                1 if self.blank else 0,
            ]
        )

    @classmethod
    def region_from_array(cls, region_bounds):
        width = int(region_bounds[2]) - region_bounds[0]
        height = int(region_bounds[3]) - region_bounds[1]
        height = np.uint8(max(height, 0))
        width = np.uint8(max(width, 0))
        frame_number = None
        if len(region_bounds) > 4:
            frame_number = region_bounds[4]
        mass = 0
        if len(region_bounds) > 5:
            mass = region_bounds[5]
        blank = False
        if len(region_bounds) > 6:
            blank = region_bounds[6] == 1
        centroid = [
            int(region_bounds[0] + width / 2),
            int(region_bounds[1] + height / 2),
        ]
        return cls(
            region_bounds[0],
            region_bounds[1],
            width,
            height,
            frame_number=np.uint16(frame_number) if frame_number is not None else None,
            mass=mass,
            blank=blank,
            centroid=centroid,
        )

    @classmethod
    def region_from_json(cls, region_json):
        frame = region_json.get("frame_number")
        if frame is None:
            frame = region_json.get("frameNumber")
        if frame is None:
            frame = region_json.get("order")
        if "centroid" in region_json:
            centroid = region_json["centroid"]
        else:
            centroid = [
                int(region_json["x"] + region_json["width"] / 2),
                int(region_json["y"] + region_json["height"] / 2),
            ]
        return cls(
            region_json["x"],
            region_json["y"],
            region_json["width"],
            region_json["height"],
            frame_number=frame,
            mass=region_json.get("mass", 0),
            blank=region_json.get("blank", False),
            pixel_variance=region_json.get("pixel_variance", 0),
            centroid=centroid,
        )

    @staticmethod
    def from_ltrb(left, top, right, bottom):
        """Construct a rectangle from left, top, right, bottom co-ords."""
        return Region(left, top, width=right - left, height=bottom - top)

    def has_moved(self, region):
        """Determines if the region has shifted horizontally or veritcally
        Not just increased in width/height
        """
        return (self.x != region.x and self.right != region.right) or (
            self.y != region.y and self.bottom != region.bottom
        )

    def calculate_variance(self, filtered, prev_filtered):
        """
        calculates variance on this frame for this region
        filtered is assumed to be cropped to the region
        """
        height, width = filtered.shape
        assert (
            width == self.width and height == self.height
        ), "calculating variance on incorrectly sized filtered"
        self.pixel_variance = tools.calculate_variance(filtered, prev_filtered)

    def set_is_along_border(self, bounds, edge=0):
        self.is_along_border = (
            self.was_cropped
            or self.x <= bounds.x + edge
            or self.y <= bounds.y + edge
            or self.right >= bounds.width - edge
            or self.bottom >= bounds.height - edge
        )

    def copy(self):
        return Region(
            self.x,
            self.y,
            self.width,
            self.height,
            self.centroid,
            self.mass,
            self.frame_number,
            self.pixel_variance,
            self.id,
            self.was_cropped,
            self.blank,
            self.is_along_border,
        )

    def average_distance(self, other):
        """Calculates the distance between 2 regions by using the distance between
        (top, left), mid points and (bottom,right) of each region
        """
        distances = []

        expected_x = int(other.x)
        expected_y = int(other.y)
        distance = tools.eucl_distance_sq((expected_x, expected_y), (self.x, self.y))
        distances.append(distance)

        expected_x = int(other.mid_x)
        expected_y = int(other.mid_y)
        distance = tools.eucl_distance_sq(
            (expected_x, expected_y), (self.mid_x, self.mid_y)
        )
        distances.append(distance)

        distance = tools.eucl_distance_sq(
            (
                other.right,
                other.bottom,
            ),
            (self.right, self.bottom),
        )
        # expected_x = int(other.right)
        # expected_y = int(other.bottom)
        # distance = tools.eucl_distance_sq((expected_x, expected_y), (self.x, self.y))
        distances.append(distance)

        return distances

    def on_height_edge(self, crop_region):
        if self.top == crop_region.top or self.bottom == crop_region.bottom:
            return True
        return False

    def on_width_edge(self, crop_region):
        if self.left == crop_region.left or self.right == crop_region.right:
            return True
        return False

    def calculate_mass(self, filtered, threshold):
        """
        calculates mass on this frame for this region
        filtered is assumed to be cropped to the region
        """
        height, width = filtered.shape
        assert (
            width == self.width and height == self.height
        ), "calculating variance on incorrectly sized filtered"

        self.mass = calculate_mass(filtered, threshold)


def calculate_mass(filtered, threshold):
    """Calculates mass of filtered frame with threshold applied"""
    if filtered.size == 0:
        return 0
    _, mass = blur_and_return_as_mask(filtered, threshold=threshold)
    return np.uint16(mass)


def blur_and_return_as_mask(frame, threshold):
    """
    Creates a binary mask out of an image by applying a threshold.
    Any pixels more than the threshold are set 1, all others are set to 0.
    A blur is also applied as a filtering step
    """
    import cv2

    thresh = cv2.GaussianBlur(frame, (5, 5), 0)
    thresh[thresh - threshold < 0] = 0
    values = thresh[thresh > 0]
    mass = len(values)
    values = 1
    return thresh, mass



import attr
import yaml
from ml_tools.rectangle import Rectangle


@attr.s
class TestConfig(yaml.YAMLObject):
    yaml_tag = "!TestConfig"
    clip_dir = attr.ib()
    recording_tests = attr.ib()
    server = attr.ib()

    @classmethod
    def load_from_file(cls, filename):
        with open(filename) as stream:
            tests = yaml.load(stream, Loader=yaml.Loader)
        for test in tests.recording_tests:
            for track in test.tracks:
                track.start_pos = Rectangle.from_ltrb(*track.start_pos[1])
                track.end_pos = Rectangle.from_ltrb(*track.end_pos[1])
        return tests


@attr.s
class TestRecording(yaml.YAMLObject):
    yaml_tag = "!TestRecording"
    rec_id = attr.ib()
    filename = attr.ib()
    device_id = attr.ib()
    device = attr.ib()
    group_id = attr.ib()
    group = attr.ib()
    tracks = attr.ib()

    @classmethod
    def load_from_meta(cls, rec_meta, track_meta, filepath):
        rec_id = rec_meta["id"]
        filename = filepath.name
        device_id = rec_meta["Device"]["id"]
        device = rec_meta["Device"]["devicename"]
        group_id = rec_meta["GroupId"]
        group = rec_meta["Group"]["groupname"]
        tracks = []

        for track in track_meta["tracks"]:
            tag = get_best_tag(track)
            if tag is None:
                tag = {}
            test_track = TestTrack.load_from_meta(rec_id, track, tag)
            tracks.append(test_track)
        return cls(rec_id, filename, device_id, device, group_id, group, tracks)


@attr.s(eq=False)
class TestTrack(yaml.YAMLObject):
    yaml_tag = "!TestTrack"
    id = attr.ib()
    track_id = attr.ib()
    tag = attr.ib()
    start = attr.ib()
    end = attr.ib()
    opt_start = attr.ib()
    opt_end = attr.ib()
    start_pos = attr.ib()
    end_pos = attr.ib()
    confidence = attr.ib()
    expected = attr.ib()

    @classmethod
    def load_from_meta(cls, rec_id, track, tag):
        start = track["data"]["start_s"]
        end = track["data"]["end_s"]
        return cls(
            id=rec_id,
            track_id=track["id"],
            tag=tag.get("what"),
            start=start,
            end=end,
            opt_start=start,
            opt_end=end,
            start_pos=track["data"]["positions"][0],
            end_pos=track["data"]["positions"][-1],
            confidence=tag.get("confidence", None),
            expected=True,
        )

    def calc_error(self):
        score = self.opt_start - self.start
        score += self.opt_end - self.end
        score = round(score, 1)
        return score

    def opt_length(self):
        return round(self.opt_end - self.opt_start, 1)

    def length(self):
        return round(self.end - self.start, 1)


def get_best_tag(track, min_confidence=0.6):
    """returns highest precidence tag from the metadata"""

    track_tags = track.get("TrackTags", [])
    track_tags = [tag for tag in track_tags if tag.get("confidence") > min_confidence]

    if not track_tags:
        return None

    # sort by original model so it has first pick if confidence is the same
    track_tags = sorted(
        track_tags,
        key=lambda x: 0 if model_name(x) == "Original" else 1,
    )

    track_tags = sorted(track_tags, key=lambda x: -1 * x["confidence"])
    manual_tags = [tag for tag in track_tags if not tag.get("automatic", False)]
    if len(manual_tags) > 0:
        return manual_tags[0]

    return track_tags[0]


def model_name(x):
    data = x.get("data")
    if not data:
        return "Original"
    return data.get("name", "Original")



import argparse
import os
import cv2
from pathlib import Path
import numpy as np
import json
import time


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-m",
        "--model",
        help="open vino model to use",
    )
    parser.add_argument(
        "-d",
        "--device",
        default="CPU",
        help="Open vino device to use (CPU/MYRIAD)",
    )
    parser.add_argument(
        "source",
        help="an image to classify",
    )

    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    print(args)
    image = cv2.imread(args.source)
    new_image = np.zeros((160, 160, 3))
    print("image is", image.shape)
    new_image[:, :, 0] = image[:, :, 2]
    new_image[:, :, 1] = image[:, :, 2]
    new_image[:, :, 2] = image[:, :, 0]
    new_image = image
    new_image = new_image / 128 - 1
    print(new_image.dtype, np.amin(new_image), np.amax(new_image))
    # preprocessing
    interpreter = NeuralInterpreter(args.model, args.device)
    results = interpreter.predict(new_image)
    max_i = np.argmax(results)
    print("Predicted results are ", np.round(100 * results), interpreter.labels[max_i])


class NeuralInterpreter:
    def __init__(self, model_name, device="CPU"):
        from openvino.inference_engine import IENetwork, IECore

        # super().__init__(model_name)
        model_name = Path(model_name)
        # can use to test on PC
        # device = "CPU"
        # device = "MYRIAD"
        model_meta = model_name.with_suffix(".txt")
        stats = json.load(open(model_meta, "r"))
        self.labels = stats.get("labels")
        model_xml = model_name.with_suffix(".xml")
        model_bin = model_name.with_suffix(".bin")
        ie = IECore()
        ie.set_config({}, device)
        net = ie.read_network(model=model_xml, weights=model_bin)
        self.input_blob = next(iter(net.input_info))
        self.out_blob = next(iter(net.outputs))
        net.batch_size = 1
        self.input_shape = net.input_info[self.input_blob].input_data.shape

        self.exec_net = ie.load_network(network=net, device_name=device)

    def predict(self, input_x):
        if input_x is None:
            return None
        input_x = np.float16(input_x)
        # rearranged_arr = np.transpose(input_x, axes=[2, 0, 1])
        input_x = np.array([[input_x]])
        s = time.time()
        res = self.exec_net.infer(inputs={self.input_blob: input_x})

        res = res[self.out_blob]
        print("time take is", time.time() - s)
        return res[0]

    def shape(self):
        return self.input_shape


if __name__ == "__main__":
    main()



import attr
import absl.logging
import argparse
import os
import logging
import math
import sys
from enum import Enum
from pathlib import Path
from cacophonyapi.user import UserAPI

from ml_tools import tools
from config.config import Config

from track.trackextractor import TrackExtractor
from .testconfig import TestConfig

MATCH_ERROR = 1


@attr.s
class Summary:
    better_tracking = attr.ib(default=0)
    same_tracking = attr.ib(default=0)
    worse_tracking = attr.ib(default=0)
    classify_incorrect = attr.ib(default=0)
    classified_correct = attr.ib(default=0)
    total_tests = attr.ib(default=0)
    unmatched_tests = attr.ib(default=0)
    unmatched_tracks = attr.ib(default=0)

    def update(self, other):
        self.better_tracking += other.better_tracking
        self.same_tracking += other.same_tracking
        self.worse_tracking += other.worse_tracking
        self.classify_incorrect += other.classify_incorrect
        self.classified_correct += other.classified_correct
        self.total_tests += other.total_tests
        self.unmatched_tests += other.unmatched_tests
        self.unmatched_tracks += other.unmatched_tracks

    @property
    def classified_percentage(self):
        if self.total_tests == 0:
            return 0
        return round(100.0 * self.classified_correct / self.total_tests)

    @property
    def tracked_well_percentage(self):
        if self.total_tests == 0:
            return 0
        return round(
            100.0 * (self.same_tracking + self.better_tracking) / self.total_tests
        )

    def print_summary(self):
        print("===== OVERAL =====")
        print(
            "Classify Results {}% {}/{}".format(
                self.classified_percentage,
                self.classified_correct,
                self.total_tests,
            )
        )
        print(
            "Tracking Results Better/Same {}% {}/{} With {} unmatched tracks (false-positives) and {} missed tests".format(
                self.tracked_well_percentage,
                self.same_tracking + self.better_tracking,
                self.total_tests,
                self.unmatched_tracks,
                self.unmatched_tests,
            )
        )


class TrackingStatus(Enum):
    IMPROVED = 1
    SAME = 0
    WORSE = -1


def match_track(gen_track, expected_tracks):
    score = None
    match = None
    MAX_ERROR = 8
    for track in expected_tracks:
        start_diff = abs(track.start - gen_track.start_s)

        gen_start = gen_track.bounds_history[0]
        distance = tools.eucl_distance_sq(
            (track.start_pos.mid_x, track.start_pos.mid_y),
            (gen_start.mid_x, gen_start.mid_y),
        )
        distance += tools.eucl_distance_sq(
            (track.start_pos.x, track.start_pos.y), (gen_start.x, gen_start.y)
        )
        distance += tools.eucl_distance_sq(
            (track.start_pos.right, track.start_pos.bottom),
            (gen_start.right, gen_start.bottom),
        )
        distance /= 3.0
        distance = math.sqrt(distance)

        # makes it more comparable to start error
        distance /= 4.0
        new_score = distance + start_diff
        if new_score > MAX_ERROR:
            continue
        if score is None or new_score < score:
            match = track
            score = new_score
    return match


class RecordingMatch:
    def __init__(self, filename, id_):
        self.matches = []
        self.unmatched_tracks = []
        self.unmatched_tests = []
        self.filename = filename
        self.id = id_
        self.summary = Summary()

    def match(self, test, tracks, predictions=None):
        self.summary.total_tests += len(test.tracks)
        gen_tracks = sorted(tracks, key=lambda x: x.get_id())
        gen_tracks = sorted(gen_tracks, key=lambda x: x.start_s)

        self.unmatched_tests = set(test.tracks)
        predicted_tag = None
        for i, track in enumerate(gen_tracks):
            if predictions is not None:
                prediction = predictions.prediction_for(track.get_id())
                predicted_tag = prediction.predicted_tag()
            test_track = match_track(track, self.unmatched_tests)
            if test_track is not None:
                self.unmatched_tests.remove(test_track)
                match = Match(test_track, track, predicted_tag)
                self.new_match(match)
            else:
                self.unmatched_tracks.append((predicted_tag, track))
                self.summary.unmatched_tracks += 1
                print(
                    "Unmatched track start {} end {}".format(
                        track.start_s,
                        track.end_s,
                    )
                )
        self.summary.unmatched_tests = len(self.unmatched_tests)

    def new_match(self, match):
        if match.status == TrackingStatus.IMPROVED:
            self.summary.better_tracking += 1
        elif match.status == TrackingStatus.SAME:
            self.summary.same_tracking += 1
        else:
            self.summary.worse_tracking += 1
        if match.tag_match():
            self.summary.classified_correct += 1
        else:
            self.summary.classify_incorrect += 1
        self.matches.append(match)

    def print_summary(self):
        print("*******{} Classification Results ******".format(self.id))
        print(
            "matches {}\tmismatches {}\tunmatched {}".format(
                self.summary.classified_correct,
                self.summary.classify_incorrect,
                self.summary.unmatched_tracks,
            )
        )
        print("*******Tracking******")
        print(
            "same {} better {}\t worse {}".format(
                self.summary.same_tracking,
                self.summary.better_tracking,
                self.summary.worse_tracking,
            )
        )
        if self.summary.unmatched_tests > 0:
            print("unmatched tests {}\t ".format(self.summary.unmatched_tests))

    def write_results(self, f):
        f.write("{}{}{}\n".format("-" * 10, "Recording", "-" * 90))
        f.write("Recordings[{}] {}\n\n".format(self.id, self.filename))
        for match in self.matches:
            match.write_results(f)

        if self.summary.unmatched_tracks > 0:
            f.write("Unmatched Tracks\n")
        for what, track in self.unmatched_tracks:
            f.write(
                "{} - [{}s]Start-End {} - {}\n".format(
                    what,
                    round(track.end_s - track.start_s, 1),
                    round(track.start_s, 1),
                    round(track.end_s, 1),
                )
            )
        f.write("\n")

        if self.summary.unmatched_tests > 0:
            f.write("Unmatched Tests\n")
        for expected in self.unmatched_tests:
            f.write(
                "{} - Opt[{}s] Start-End {} - {}, Expected[{}s] {} - {}\n".format(
                    expected.tag,
                    round(expected.opt_end - expected.opt_start, 1),
                    expected.opt_start,
                    expected.opt_end,
                    round(expected.end - expected.start, 1),
                    expected.start,
                    expected.end,
                )
            )
        f.write("\n")


class Match:
    def __init__(self, test_track, track, tag=None):
        expected_length = test_track.opt_end - test_track.opt_start
        self.length_diff = round(expected_length - (track.end_s - track.start_s), 2)
        self.start_diff_s = round(test_track.start - track.start_s, 2)
        self.end_diff_s = round(test_track.end - track.end_s, 2)
        self.opt_start_diff_s = round(test_track.opt_start - track.start_s, 2)
        self.opt_end_diff_s = round(test_track.opt_end - track.end_s, 2)
        self.error = round(abs(self.opt_start_diff_s) + abs(self.opt_end_diff_s), 1)

        if self.error <= test_track.calc_error():
            self.status = TrackingStatus.IMPROVED
        elif self.error < MATCH_ERROR:
            self.status = TrackingStatus.SAME
        else:
            self.status = TrackingStatus.WORSE
        self.expected_tag = test_track.tag
        self.got_animal = tag
        self.test_track = test_track
        self.track = track

    def tracking_status(self):
        if self.status == TrackingStatus.IMPROVED:
            return "Better Tracking"
        elif self.status == TrackingStatus.SAME:
            return "Same Tracking"
        return "Worse Tracking"

    def classify_status(self):
        if self.expected_tag == self.got_animal:
            return "Classified Correctly"
        return "Classified Incorrect"

    def write_results(self, f):
        f.write("{}{}{}\n".format("=" * 10, "Track", "=" * 90))

        f.write("{}\t{}\n".format(self.tracking_status(), self.classify_status()))
        f.write("Exepcted:\n")
        f.write(
            "{} - Opt[{}s] Start-End {} - {}, Expected[{}s] {} - {}\n".format(
                self.expected_tag,
                self.test_track.opt_length(),
                self.test_track.opt_start,
                self.test_track.opt_end,
                self.test_track.length(),
                self.test_track.start,
                self.test_track.end,
            )
        )
        f.write("Got:\n")
        f.write(
            "{} - [{}s]Start-End {} - {}\n".format(
                self.got_animal,
                round(self.track.end_s - self.track.start_s, 1),
                round(self.track.start_s, 1),
                round(self.track.end_s, 1),
            )
        )
        f.write("\n")

    def tag_match(self):
        return self.expected_tag == self.got_animal


class TestClassify:
    def __init__(self, args):
        self.test_config = TestConfig.load_from_file(args.tests)
        self.classifier_config = Config.load_from_file(args.classify_config)
        if args.model_file:
            model_file = args.model_file
        self.track_extractor = TrackExtractor(self.classifier_config)
        # try download missing tests
        if args.user and args.password:
            api = UserAPI(args.server, args.user, args.password)
            out_dir = Path(self.test_config.clip_dir)
            if not out_dir.exists():
                out_dir.mkdir()
            for test in self.test_config.recording_tests:
                filepath = out_dir / test.filename
                if not filepath.exists():
                    if iter_to_file(
                        out_dir / test.filename, api.download_raw(test.rec_id)
                    ):
                        logging.info("Saved %s", filepath)
        self.results = []

    def run_tests(self, args):
        out_dir = Path(self.test_config.clip_dir)

        for test in self.test_config.recording_tests:
            filepath = out_dir / test.filename
            if not filepath.exists():
                logging.info(
                    " %s not found, add cmd args user and password to download this file from the cacophony server",
                    filepath,
                )
                continue
            logging.info("testing %s ", test.filename)
            clip = self.track_extractor.extract_tracks(filepath)
            meta_filename = filepath.with_suffix(".txt")
            self.track_extractor.save_metadata(str(filepath), meta_filename, clip, 0)

            predictions = None
            rec_match = self.compare_output(clip, predictions, test)
            if self.track_extractor.previewer:
                mpeg_filename = filepath.with_suffix(".mp4")
                logging.info("Exporting preview to '%s'", mpeg_filename)
                self.track_extractor.previewer.export_clip_preview(
                    mpeg_filename, clip, predictions
                )
            self.results.append(rec_match)

    def write_results(self, filename, config_file):
        with open(filename, "w") as f:
            for res in self.results:
                res.write_results(f)
            f.write("Config\n")
            with open(config_file, "r") as config:
                f.write(config.read())

    def print_summary(self):
        print("===== SUMMARY =====")
        total_summary = Summary()
        for result in self.results:
            result.print_summary()
            total_summary.update(result.summary)

        total_summary.print_summary()

    def compare_output(self, clip, predictions, expected):
        rec_match = RecordingMatch(clip.source_file, expected.rec_id)
        rec_match.match(expected, clip.tracks, predictions)
        return rec_match


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--classify_config",
        default="./classifier.yaml",
        help="config file to classify with",
    )

    parser.add_argument(
        "-t",
        "--tests",
        default="tests/tracking-tests.yml",
        help="YML file containing tests",
    )

    parser.add_argument(
        "-o",
        "--output",
        default="tests/tracking-results.txt",
        help="Text file descriibing the results",
    )

    parser.add_argument(
        "-m",
        "--model-file",
        help="Path to model file to use, will override config model",
    )

    parser.add_argument("--user", help="API server username")
    parser.add_argument("--password", help="API server password")
    parser.add_argument(
        "-s",
        "--server",
        default="https://api.cacophony.org.nz",
        help="CPTV file server URL",
    )
    args = parser.parse_args()
    return args


def init_logging():
    logging.root.removeHandler(absl.logging._absl_handler)
    absl.logging._warn_preinit_stderr = False
    logging.basicConfig(stream=sys.stderr, level=logging.INFO)


def main():
    init_logging()
    args = parse_args()
    test = TestClassify(args)
    test.run_tests(args)
    test.write_results(args.output, args.classify_config)
    test.print_summary()


if __name__ == "__main__":
    main()


def iter_to_file(filename, source, overwrite=True):
    if not overwrite and Path(filename).is_file():
        print("{} already exists".format(filename))
        return False
    with open(filename, "wb") as f:
        for chunk in source:
            f.write(chunk)
    return True



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from PIL import Image
from pathlib import Path
from multiprocessing import Process, Queue

import collections
import hashlib
import io
import json
import multiprocessing
import os
import time
from absl import app
from absl import flags
from absl import logging
import numpy as np
from PIL import Image, ImageOps

import tensorflow as tf
from . import tfrecord_util
from ml_tools import tools
from ml_tools.imageprocessing import normalize, rotate
from track.cliptracker import get_diff_back_filtered
import cv2
import random
import math


def create_tf_example(sample, thermal, filtered, labels):
    """Converts image and annotations to a tf.Example proto.

    Args:
      image: dict with keys: [u'license', u'file_name', u'coco_url', u'height',
        u'width', u'date_captured', u'flickr_url', u'id']
      image_dir: directory containing the image files.
      bbox_annotations:
        list of dicts with keys: [u'segmentation', u'area', u'iscrowd',
          u'image_id', u'bbox', u'category_id', u'id'] Notice that bounding box
          coordinates in the official COCO dataset are given as [x, y, width,
          height] tuples using absolute coordinates where x, y represent the
          top-left (0-indexed) corner.  This function converts to the format
          expected by the Tensorflow Object Detection API (which is which is
          [ymin, xmin, ymax, xmax] with coordinates normalized relative to image
          size).
      category_index: a dict containing COCO category information keyed by the
        'id' field of each category.  See the label_map_util.create_category_index
        function.
      caption_annotations:
        list of dict with keys: [u'id', u'image_id', u'str'].
      include_masks: Whether to include instance segmentations masks
        (PNG encoded) in the result. default: False.

    Returns:
      example: The converted tf.Example
      num_annotations_skipped: Number of (invalid) annotations that were ignored.

    Raises:
      ValueError: if the image pointed to by data['filename'] is not a valid JPEG
    """
    image_height, image_width = thermal.shape

    image = Image.fromarray(np.uint8(thermal))

    image_id = sample.unique_id

    encoded_jpg_io = io.BytesIO()
    image.save(encoded_jpg_io, format="PNG", quality=100, subsampling=0)

    encoded_thermal = encoded_jpg_io.getvalue()
    thermal_key = hashlib.sha256(encoded_thermal).hexdigest()

    image = Image.fromarray(np.uint8(filtered))

    encoded_jpg_io = io.BytesIO()
    image.save(encoded_jpg_io, format="PNG", quality=100, subsampling=0)
    encoded_filtered = encoded_jpg_io.getvalue()
    filtered_key = hashlib.sha256(encoded_filtered).hexdigest()

    feature_dict = {
        "image/augmented": tfrecord_util.int64_feature(sample.augment),
        "image/height": tfrecord_util.int64_feature(image_height),
        "image/width": tfrecord_util.int64_feature(image_width),
        "image/filename": tfrecord_util.bytes_feature(
            str(sample.source_file).encode("utf8")
        ),
        "image/source_id": tfrecord_util.bytes_feature(str(image_id).encode("utf8")),
        "image/thermalkey/sha256": tfrecord_util.bytes_feature(
            thermal_key.encode("utf8")
        ),
        "image/thermalencoded": tfrecord_util.bytes_feature(encoded_thermal),
        "image/filteredkey/sha256": tfrecord_util.bytes_feature(
            filtered_key.encode("utf8")
        ),
        "image/clip_id": tfrecord_util.int64_feature(sample.clip_id),
        "image/track_id": tfrecord_util.int64_feature(sample.track_id),
        "image/filteredencoded": tfrecord_util.bytes_feature(encoded_filtered),
        "image/format": tfrecord_util.bytes_feature("jpeg".encode("utf8")),
        "image/class/text": tfrecord_util.bytes_feature(sample.label.encode("utf8")),
        "image/class/label": tfrecord_util.int64_feature(labels.index(sample.label)),
    }
    example = tf.train.Example(features=tf.train.Features(feature=feature_dict))
    return example


def get_data(samples, back_thresh):
    vidcap = cv2.VideoCapture(str(samples[0].source_file))
    frames = {}
    backgorund = None
    frame_num = 0
    frames_needed = [s.region.frame_number for s in samples]
    frames_needed.sort()
    if len(frames_needed) == 0:
        return []
    while True:
        for _ in range(2):
            # try read first frame twice
            success, image = vidcap.read()
            is_background_frame = False
            if success or frame_num > 0:
                break
        if not success:
            break
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        if backgorund is None:
            is_background_frame = np.all(image[:, :, 0] == image[:, :, 1]) and np.all(
                image[:, :, 1] == image[:, :, 2]
            )
            background = np.uint8(gray)
        if not is_background_frame and frame_num in frames_needed:
            frames[frame_num] = gray
            # append(gray)
        frame_num += 1
        if frame_num > frames_needed[-1]:
            break
    data = []
    failed = []
    for sample in samples:
        if sample.region.frame_number not in frames:
            failed.append(sample.region.frame_number)
            continue
        frame = frames[sample.region.frame_number]
        gray_sub = sample.region.subimage(frame)
        back_sub = sample.region.subimage(background)
        filtered = get_diff_back_filtered(back_sub, gray_sub, back_thresh)
        gray_sub, stats = normalize(gray_sub, new_max=255)
        if not stats[0]:
            continue
        filtered, stats = normalize(filtered, new_max=255)
        if not stats[0]:
            continue
        data.append((sample, gray_sub, filtered))
    if len(failed) > 0:
        logging.warning("Could not get %s for %s", failed, str(samples[0].source_file))

    return data


def save_data(samples, writer, labels, extra_args):
    sample_data = get_data(samples, extra_args["back_thresh"])
    if sample_data is None:
        return 0
    saved = 0
    try:
        for sample, thermal, filtered in sample_data:
            tf_example = create_tf_example(
                sample,
                thermal,
                filtered,
                labels,
            )
            writer.write(tf_example.SerializeToString())
            saved += 1
    except:
        logging.error(
            "Could not save data for %s", samples[0].source_file, exc_info=True
        )
    return saved



import numpy as np
import random
from ml_tools import tools
from ml_tools.frame import TrackChannels
import logging
from ml_tools import imageprocessing
from track.region import Region
import cv2
from ml_tools.tools import FrameTypes

# size to scale each frame to when loaded.

MIN_SIZE = 4
EDGE = 1

res_x = 120
res_y = 160


# this is from tf source code same as preprocess_input
def preprocess_fn(x):
    x /= 127.5
    x -= 1.0
    return x


def convert(image):
    import tensorflow as tf

    image = tf.image.convert_image_dtype(image, tf.float32)
    return image


def augement_frame(frame, frame_size, dim):
    frame = imageprocessing.resize_cv(
        frame,
        dim,
        extra_h=random.randint(0, int(frame_size * 0.05)),
        extra_v=random.randint(0, int(frame_size * 0.05)),
    )

    image = convert(frame)
    import tensorflow as tf

    image = tf.image.random_crop(image, size=[dim[0], dim[1], 3])
    if random.random() > 0.50:
        image = tf.image.flip_left_right(image)

    if random.random() > 0.20:
        image = tf.image.random_contrast(image, 0.8, 1.2)
    image = tf.minimum(image, 1.0)
    image = tf.maximum(image, 0.0)
    return image.numpy()


def preprocess_frame(
    frame,
    out_dim,
    region,
    background=None,
    crop_rectangle=None,
    calculate_filtered=True,
    filtered_norm_limits=None,
    thermal_norm_limits=None,
):
    median = np.median(frame.thermal)
    cropped_frame = frame.crop_by_region(region, only_thermal=calculate_filtered)
    cropped_frame.thermal = np.float32(cropped_frame.thermal)
    if calculate_filtered:
        if background is None:
            logging.warning(
                "Not calculating filtered frame as no background was supplied"
            )
        else:
            cropped_frame.filtered = cropped_frame.thermal - region.subimage(background)

    cropped_frame.resize_with_aspect(
        out_dim,
        crop_rectangle,
        True,
    )
    cropped_frame.thermal -= median
    if thermal_norm_limits is None:
        np.clip(cropped_frame.thermal, 0, None, out=cropped_frame.thermal)
    if filtered_norm_limits is not None:
        cropped_frame.filtered, stats = imageprocessing.normalize(
            cropped_frame.filtered,
            min=filtered_norm_limits[0],
            max=filtered_norm_limits[1],
            new_max=255,
        )
        if frame.thermal is not None:
            thermal_min = None
            thermal_max = None
            if thermal_norm_limits is not None:
                thermal_min, thermal_max = thermal_norm_limits
            cropped_frame.thermal, _ = imageprocessing.normalize(
                cropped_frame.thermal, min=thermal_min, max=thermal_max, new_max=255
            )
    else:
        cropped_frame.normalize()
    return cropped_frame


index = 0


def preprocess_single_frame(
    preprocessed_frame,
    channels,
    preprocess_fn=None,
    save_info="",
):
    data = []
    for channel in channels:
        if isinstance(channel, str):
            channel = TrackChannels[channel]
        data.append(preprocessed_frame.get_channel(channel))

    image = np.stack(
        data,
        axis=2,
    )
    # global index
    # index += 1
    # tools.saveclassify_image(
    #     image,
    #     f"samples/{save_info}-{index}",
    # )

    if preprocess_fn:
        image = preprocess_fn(image)
    return image


# index = 0


#
def preprocess_movement(
    preprocess_frames,
    frames_per_row,
    frame_size,
    channels,
    preprocess_fn=None,
    sample=None,
):
    frame_types = {}
    data = []
    frame_samples = list(np.arange(len(preprocess_frames)))
    if len(preprocess_frames) < frames_per_row * 5:
        extra_samples = np.random.choice(
            frame_samples, frames_per_row * 5 - len(preprocess_frames)
        )
        frame_samples.extend(extra_samples)
        frame_samples.sort()
    for channel in channels:
        if isinstance(channel, str):
            channel = TrackChannels[channel]
        if channel in frame_types:
            data.append(frame_types[channel])
            continue
        channel_segment = [frame.get_channel(channel) for frame in preprocess_frames]
        channel_data, success = imageprocessing.square_clip(
            channel_segment,
            frames_per_row,
            (frame_size, frame_size),
            frame_samples,
            normalize=False,
        )
        # already done normalization

        if not success:
            return None
        data.append(channel_data)
        frame_types[channel] = channel_data
    data = np.stack(data, axis=2)

    #
    # # # # # for testing
    # global index
    # index += 1
    # tools.saveclassify_image(
    #     data,
    #     f"samples/{sample}-{index}",
    # )

    if preprocess_fn:
        data = preprocess_fn(data)
    return np.float32(data)



from ml_tools.frame import TrackChannels
from ml_tools.datasetstructures import SegmentType
from ml_tools.preprocess import FrameTypes


class HyperParams(dict):
    """Helper wrapper for dictionary to make accessing hyper parameters easier"""

    def __init__(self, *args):
        super(HyperParams, self).__init__(*args)

        self.insert_defaults()

    def insert_defaults(self):
        self["model_name"] = self.model_name
        self["dense_sizes"] = self.dense_sizes
        self["base_training"] = self.base_training
        self["retrain_layer"] = self.retrain_layer
        self["dropout"] = self.dropout
        self["learning_rate"] = self.learning_rate
        self["learning_rate_decay"] = self.learning_rate_decay
        self["use_movement"] = self.use_movement
        self["use_segments"] = self.use_segments
        self["square_width"] = self.square_width
        self["frame_size"] = self.frame_size
        self["segment_width"] = self.segment_width
        self["segment_types"] = self.segment_types
        self["multi_label"] = True
        self["diff_norm"] = self.diff_norm
        self["thermal_diff_norm"] = self.thermal_diff_norm

        self["smooth_predictions"] = self.smooth_predictions
        self["channels"] = self.channels

    @property
    def channels(self):
        return self.get(
            "channels", [TrackChannels.thermal.name, TrackChannels.filtered.name]
        )

    @property
    def output_dim(self):
        if self.use_movement:
            return (
                self.frame_size * self.square_width,
                self.frame_size * self.square_width,
                len(self.channels),
            )
        return (self.frame_size, self.frame_size, len(self.channels))

    @property
    def smooth_predictions(self):
        return self.get("smooth_predictions", True)

    @property
    def excluded_labels(self):
        return self.get("excluded_labels", None)

    @property
    def remapped_labels(self):
        return self.get("remapped_labels", None)

    @property
    def thermal_diff_norm(self):
        return self.get("thermal_diff_norm", False)

    @property
    def diff_norm(self):
        return self.get("diff_norm", True)

    @property
    def multi_label(self):
        return self.get("multi_label", True)

    @property
    def keep_aspect(self):
        return self.get("keep_aspect", False)

    @property
    def use_background_filtered(self):
        return self.get("use_background_filtered", True)

    @property
    def keep_edge(self):
        return self.get("keep_edge", True)

    @property
    def segment_width(self):
        return self.get("segment_width", 25 if self.use_segments else 1)

    @property
    def segment_types(self):
        segment_types = self.get("segment_type", [SegmentType.ALL_RANDOM])
        # convert string to enum type
        if isinstance(segment_types, str):
            # old metadata
            segment_types = [SegmentType[segment_types]]
        elif isinstance(segment_types[0], str):
            for i in range(len(segment_types)):
                segment_types[i] = SegmentType[segment_types[i]]
        return segment_types

    @property
    def mvm(self):
        return self.get("mvm", False)

    @property
    def mvm_forest(self):
        return self.get("mvm_forest", False)

    @property
    def model_name(self):
        return self.get("model_name", "wr-resnet")

    @property
    def dense_sizes(self):
        return self.get("dense_sizes", None)

    @property
    def label_smoothing(self):
        return self.get("label_smoothing", 0)

    @property
    def base_training(self):
        return self.get("base_training", True)

    @property
    def retrain_layer(self):
        return self.get("retrain_layer")

    @property
    def dropout(self):
        return self.get("dropout", 0.3)

    @property
    def learning_rate(self):
        return self.get("learning_rate", 0.001)

    @property
    def learning_rate_decay(self):
        return self.get("learning_rate_decay", None)

    # Datageneration parameters
    @property
    def batch_size(self):
        return self.get("batch_size", 32)

    @property
    def lstm(self):
        return self.get("lstm", False)

    @property
    def use_movement(self):
        return self.get("use_movement", True)

    @property
    def use_segments(self):
        return self.get("use_segments", True)

    @property
    def square_width(self):
        default = 1
        if self.use_segments:
            default = 5
        return self.get("square_width", default)

    @property
    def frame_size(self):
        return self.get("frame_size", 32)

    def set_use_segments(self, use_segments):
        self["use_segments"] = use_segments
        if use_segments:
            self["square_width"] = 5
        else:
            self["square_width"] = 1

    #
    # @property
    # def red_type(self):
    #     ft = self.get("red_type", FrameTypes.thermal_tiled.name)
    #     return FrameTypes[ft]
    #
    # @property
    # def green_type(self):
    #     ft = self.get("green_type", FrameTypes.thermal_tiled.name)
    #     return FrameTypes[ft]
    #
    # @property
    # def blue_type(self):
    #     ft = self.get("blue_type", FrameTypes.thermal_tiled.name)
    #     return FrameTypes[ft]
